"""
Module for plotting atmospheric retrieval results.
"""

import sys
import warnings

from typing import Optional, Tuple

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from matplotlib.colorbar import Colorbar
from matplotlib.colors import LogNorm
from matplotlib.ticker import MultipleLocator
from scipy.interpolate import interp1d
from scipy.stats import lognorm
from typeguard import typechecked

from species.read.read_radtrans import ReadRadtrans
from species.util.retrieval_util import (
    calc_metal_ratio,
    get_condensation_curve,
    pt_ret_model,
    pt_spline_interp,
    quench_pressure,
    scale_cloud_abund,
)


@typechecked
def plot_pt_profile(
    tag: str,
    random: Optional[int] = 100,
    envelope: bool = False,
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    offset: Optional[Tuple[float, float]] = None,
    output: Optional[str] = None,
    radtrans: Optional[ReadRadtrans] = None,
    extra_axis: Optional[str] = None,
    rad_conv_bound: bool = False,
) -> mpl.figure.Figure:
    """
    Function to plot the posterior distribution.

    Parameters
    ----------
    tag : str
        Database tag with the posterior samples.
    random : int, None
        Number of randomly selected samples from the posterior. All
        samples are selected if the argument is set to ``None``.
    envelope : bool
        Plot an envelope instead of the individual samples. The
        envelopes show the 68 and 99.7 percent confidence intervals,
        so :math:`1\\sigma` and :math:`3\\sigma` in case of
        Gaussian distributions.
    xlim : tuple(float, float), None
        Limits of the temperature axis. Default values are used if
        set to ``None``.
    ylim : tuple(float, float), None
        Limits of the pressure axis. Default values are used if set
        to ``None``.
    offset : tuple(float, float), None
        Offset of the x- and y-axis label. Default values are used
        if set to ``None``.
    output : str, None
        Output filename for the plot. The plot is shown in an
        interface window if the argument is set to ``None``.
    radtrans : ReadRadtrans, None
        Instance of :class:`~species.read.read_radtrans.ReadRadtrans`.
        Not used if set to ``None``.
    extra_axis : str, None
        The quantify that is plotted at the top axis ('photosphere',
        'grains'). The top axis is not used if the argument is set
        to ``None``.
    rad_conv_bound : bool
        Plot the range of pressures (:math:`\\pm 1\\sigma`) of the
        radiative-convective boundary.

    Returns
    -------
    matplotlib.figure.Figure
        The ``Figure`` object that can be used for further
        customization of the plot.
    """

    if output is None:
        print("Plotting the P-T profiles...", end="", flush=True)
    else:
        print(f"Plotting the P-T profiles: {output}...", end="", flush=True)

    cloud_colors = [
        "tab:blue",
        "tab:orange",
        "tab:green",
        "tab:cyan",
        "tab:pink",
        "tab:brown",
        "tab:olive",
    ]

    color_iter = iter(cloud_colors)

    from species.data.database import Database

    species_db = Database()
    box = species_db.get_samples(tag)

    parameters = np.asarray(box.parameters)
    samples = box.samples
    model_param = box.prob_sample

    if random is not None:
        indices = np.random.randint(samples.shape[0], size=random)
        samples = samples[indices,]

    param_index = {}
    for item in parameters:
        param_index[item] = np.argwhere(parameters == item)[0][0]

    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "dejavuserif"

    fig = plt.figure(figsize=(4.0, 5.0))
    gridsp = mpl.gridspec.GridSpec(1, 1)
    gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

    ax = plt.subplot(gridsp[0, 0])

    ax.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax.set_xlabel("Temperature (K)", fontsize=13)
    ax.set_ylabel("Pressure (bar)", fontsize=13)

    if offset is not None:
        ax.get_xaxis().set_label_coords(0.5, offset[0])
        ax.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax.get_xaxis().set_label_coords(0.5, -0.06)
        ax.get_yaxis().set_label_coords(-0.14, 0.5)

    if "temp_nodes" in box.attributes:
        temp_nodes = box.attributes["temp_nodes"]
    else:
        # For backward compatibility
        temp_nodes = 15

    if "abund_nodes" in box.attributes:
        if box.attributes["abund_nodes"] == "None":
            abund_nodes = None
        else:
            abund_nodes = box.attributes["abund_nodes"]
    else:
        # For backward compatibility
        abund_nodes = None

    if "max_press" in box.attributes:
        max_press = box.attributes["max_press"]
    else:
        # For backward compatibility
        max_press = 1e3  # (bar)

    if xlim is None:
        ax.set_xlim(1000.0, 5000.0)
    else:
        ax.set_xlim(xlim[0], xlim[1])

    if ylim is None:
        ax.set_ylim(max_press, 1e-6)
    else:
        ax.set_ylim(ylim[0], ylim[1])

    ax.set_yscale("log")

    # Create the pressure points (bar)
    pressure = np.logspace(-6.0, np.log10(max_press), 180)

    if "tint" in parameters and "log_delta" in parameters and "alpha" in parameters:
        pt_profile = "molliere"

    elif "tint" in parameters and "log_delta" in parameters:
        pt_profile = "eddington"

    elif "T_bottom" in parameters and "PTslope_1" in parameters:
        pt_profile = "gradient"

    else:
        pt_profile = "free"

        temp_index = []
        for i in range(temp_nodes):
            temp_index.append(np.argwhere(parameters == f"t{i}")[0][0])

        knot_press = np.logspace(
            np.log10(pressure[0]), np.log10(pressure[-1]), temp_nodes
        )

    if pt_profile == "molliere":
        conv_press = np.zeros(samples.shape[0])

    if envelope:
        temp_list = np.zeros((samples.shape[0], pressure.shape[0]))
    else:
        temp_list = None

    for i, item in enumerate(samples):
        # C/O and [Fe/H]

        if box.attributes["chemistry"] == "equilibrium":
            metallicity = item[param_index["metallicity"]]
            c_o_ratio = item[param_index["c_o_ratio"]]

        elif box.attributes["chemistry"] == "free":
            # TODO Set [Fe/H] = 0
            metallicity = 0.0

            # Create a dictionary with the mass fractions

            if abund_nodes is None:
                log_x_abund = {}
                line_species = []

                for line_idx in range(box.attributes["n_line_species"]):
                    line_item = box.attributes[f"line_species{line_idx}"]
                    log_x_abund[line_item] = item[param_index[line_item]]
                    line_species.append(line_item)

                # Check if the C/H and O/H ratios are within the prior boundaries

                _, _, c_o_ratio = calc_metal_ratio(log_x_abund, line_species)

            else:
                log_x_abund = {}
                for line_idx in range(box.attributes["n_line_species"]):
                    line_item = box.attributes[f"line_species{line_idx}"]
                    for node_idx in range(abund_nodes):
                        log_x_abund[f"{line_item}_{node_idx}"] = model_param[
                            f"{line_item}_{node_idx}"
                        ]

                # TODO Set C/O = 0.55 for Molliere P-T profile
                # and cloud condensation profiles
                c_o_ratio = 0.55

        if pt_profile == "molliere":
            t3_param = np.array(
                [
                    item[param_index["t1"]],
                    item[param_index["t2"]],
                    item[param_index["t3"]],
                ]
            )

            temp, _, conv_press[i] = pt_ret_model(
                t3_param,
                10.0 ** item[param_index["log_delta"]],
                item[param_index["alpha"]],
                item[param_index["tint"]],
                pressure,
                metallicity,
                c_o_ratio,
            )

        elif pt_profile == "eddington":
            tau = pressure * 1e6 * 10.0 ** item[param_index["log_delta"]]
            temp = (0.75 * item[param_index["tint"]] ** 4.0 * (2.0 / 3.0 + tau)) ** 0.25

        elif pt_profile == "gradient":
            num_layer = 6  # could make a variable in the future
            layer_pt_slopes = np.ones(num_layer) * np.nan
            for index in range(num_layer):
                layer_pt_slopes[index] = item[
                    param_index[f"PTslope_{num_layer - index}"]
                ]

            try:
                from petitRADTRANS.physics import dTdP_temperature_profile
            except ImportError as import_error:
                raise ImportError(
                    "Can't import the dTdP profile function from "
                    "petitRADTRANS, check that your version of pRT "
                    "includes this function in petitRADTRANS.physics",
                    import_error,
                )

            temp = dTdP_temperature_profile(
                pressure,
                num_layer,  # could change in the future
                layer_pt_slopes,
                item[param_index["T_bottom"]],
            )

        elif pt_profile == "free":
            knot_temp = []
            for j in range(temp_nodes):
                knot_temp.append(item[temp_index[j]])

            knot_temp = np.asarray(knot_temp)

            if "pt_smooth" in parameters:
                pt_smooth = item[param_index["pt_smooth"]]

            elif "pt_smooth_0" in parameters:
                pt_smooth = {}
                for i in range(temp_nodes - 1):
                    pt_smooth[f"pt_smooth_{i}"] = item[param_index[f"pt_smooth_{i}"]]

            elif "pt_turn" in parameters:
                pt_smooth = {
                    "pt_smooth_1": item[param_index["pt_smooth_1"]],
                    "pt_smooth_2": item[param_index["pt_smooth_2"]],
                    "pt_turn": item[param_index["pt_turn"]],
                    "pt_index": item[param_index["pt_index"]],
                }

            else:
                pt_smooth = box.attributes["pt_smooth"]

            temp = pt_spline_interp(
                knot_press, knot_temp, pressure, pt_smooth=pt_smooth
            )

        # if pt_profile == "free":
        #     temp = temp[:, 0]
        #
        #     if "poor_mans_nonequ_chem" in sys.modules:
        #         from poor_mans_nonequ_chem.poor_mans_nonequ_chem import interpol_abundances
        #     else:
        #         from petitRADTRANS.poor_mans_nonequ_chem.poor_mans_nonequ_chem import interpol_abundances
        #
        #     ab = interpol_abundances(
        #         np.full(temp.shape[0], c_o_ratio),
        #         np.full(temp.shape[0], metallicity),
        #         temp,
        #         pressure,
        #     )
        #
        #     nabla_ad = ab["nabla_ad"]
        #
        #     # Convert pressures from bar to cgs units
        #     press_cgs = pressure * 1e6
        #
        #     # Calculate the current, radiative temperature gradient
        #     nab_rad = np.diff(np.log(temp)) / np.diff(np.log(press_cgs))
        #
        #     # Extend to array of same length as pressure structure
        #     nabla_rad = np.ones_like(temp)
        #     nabla_rad[0] = nab_rad[0]
        #     nabla_rad[-1] = nab_rad[-1]
        #     nabla_rad[1:-1] = (nab_rad[1:] + nab_rad[:-1]) / 2.0
        #
        #     # Where is the atmosphere convectively unstable?
        #     conv_index = nabla_rad > nabla_ad
        #
        #     tfinal = None
        #
        #     for i in range(10):
        #         if i == 0:
        #             t_take = copy(temp)
        #         else:
        #             t_take = copy(tfinal)
        #
        #         ab = interpol_abundances(
        #             np.full(t_take.shape[0], c_o_ratio),
        #             np.full(t_take.shape[0], metallicity),
        #             t_take,
        #             pressure,
        #         )
        #
        #         nabla_ad = ab["nabla_ad"]
        #
        #         # Calculate the average nabla_ad between the layers
        #         nabla_ad_mean = nabla_ad
        #         nabla_ad_mean[1:] = (nabla_ad[1:] + nabla_ad[:-1]) / 2.0
        #
        #         # What are the increments in temperature due to convection
        #         tnew = nabla_ad_mean[conv_index] * np.mean(np.diff(np.log(press_cgs)))
        #
        #         # What is the last radiative temperature?
        #         tstart = np.log(t_take[~conv_index][-1])
        #
        #         # Integrate and translate to temperature
        #         # from log(temperature)
        #         tnew = np.exp(np.cumsum(tnew) + tstart)
        #
        #         # Add upper radiative and lower covective
        #         # part into one single array
        #         tfinal = copy(t_take)
        #         tfinal[conv_index] = tnew
        #
        #         if np.max(np.abs(t_take - tfinal) / t_take) < 0.01:
        #             break
        #
        #     temp = copy(tfinal)

        if envelope:
            temp_list[i] = temp
        else:
            ax.plot(temp, pressure, "-", lw=0.3, color="gray", zorder=1)

    if box.attributes["chemistry"] == "free":
        # TODO Set [Fe/H] = 0
        model_param["metallicity"] = metallicity
        model_param["c_o_ratio"] = c_o_ratio

    if pt_profile == "molliere":
        temp, _, conv_press_median = pt_ret_model(
            np.array([model_param["t1"], model_param["t2"], model_param["t3"]]),
            10.0 ** model_param["log_delta"],
            model_param["alpha"],
            model_param["tint"],
            pressure,
            model_param["metallicity"],
            model_param["c_o_ratio"],
        )

        if rad_conv_bound:
            press_min = np.mean(conv_press) - np.std(conv_press)
            press_max = np.mean(conv_press) + np.std(conv_press)

            ax.axhspan(
                press_min,
                press_max,
                zorder=0,
                color="lightsteelblue",
                linewidth=0.0,
                alpha=0.5,
            )

            ax.axhline(conv_press_median, zorder=0, color="cornflowerblue", alpha=0.5)

    elif pt_profile == "eddington":
        tau = pressure * 1e6 * 10.0 ** model_param["log_delta"]
        temp = (0.75 * model_param["tint"] ** 4.0 * (2.0 / 3.0 + tau)) ** 0.25

    elif pt_profile == "free":
        knot_temp = []
        for i in range(temp_nodes):
            knot_temp.append(model_param[f"t{i}"])

        knot_temp = np.asarray(knot_temp)

        ax.plot(knot_temp, knot_press, "o", ms=5.0, mew=0.0, color="tomato", zorder=3.0)

        if "pt_smooth" in parameters:
            pt_smooth = model_param["pt_smooth"]

        elif "pt_smooth_0" in parameters:
            pt_smooth = {}
            for i in range(temp_nodes - 1):
                pt_smooth[f"pt_smooth_{i}"] = item[param_index[f"pt_smooth_{i}"]]

        elif "pt_turn" in parameters:
            pt_smooth = {
                "pt_smooth_1": model_param["pt_smooth_1"],
                "pt_smooth_2": model_param["pt_smooth_2"],
                "pt_turn": model_param["pt_turn"],
                "pt_index": model_param["pt_index"],
            }

        else:
            pt_smooth = box.attributes["pt_smooth"]

        temp = pt_spline_interp(knot_press, knot_temp, pressure, pt_smooth=pt_smooth)

    if envelope:
        temp_percent = np.percentile(temp_list, [0.3, 16.0, 84.0, 99.7], axis=0)

        ax.fill_betweenx(
            y=pressure,
            x1=temp_percent[0],
            x2=temp_percent[3],
            color="peachpuff",
            alpha=0.4,
            zorder=1,
            linewidth=0.0,
        )

        ax.fill_betweenx(
            y=pressure,
            x1=temp_percent[1],
            x2=temp_percent[2],
            color="peachpuff",
            alpha=1.0,
            zorder=1,
            linewidth=0.0,
        )

    ax.plot(temp, pressure, "-", lw=1, color="black", zorder=2)

    # data = np.loadtxt('res_struct.dat')
    # ax.plot(data[:, 1], data[:, 0], lw=1, color='tab:purple')

    # Add cloud condensation profiles

    if (
        extra_axis == "grains"
        and "metallicity" in model_param
        and "c_o_ratio" in model_param
    ):
        if box.attributes["quenching"] == "pressure":
            p_quench = 10.0 ** model_param["log_p_quench"]

        elif box.attributes["quenching"] == "diffusion":
            p_quench = quench_pressure(
                radtrans.rt_object.pressures,
                radtrans.rt_object.temperatures,
                model_param["metallicity"],
                model_param["c_o_ratio"],
                model_param["logg"],
                model_param["log_kzz"],
            )

        else:
            p_quench = None

        # Import interpol_abundances here because it is slow

        from petitRADTRANS.chemistry.pre_calculated_chemistry import (
            PreCalculatedEquilibriumChemistryTable,
        )

        eq_chem = PreCalculatedEquilibriumChemistryTable()

        abund_in, mmw, _ = eq_chem.interpolate_mass_fractions(
            np.full(pressure.size, model_param["c_o_ratio"]),
            np.full(pressure.size, model_param["metallicity"]),
            temp,
            pressure,
            carbon_pressure_quench=p_quench,
            full=True,
        )

        for cloud_item in radtrans.cloud_species:
            if f"{cloud_item}_tau" in model_param:
                # Calculate the scaled mass fraction of the clouds
                model_param[f"{cloud_item}_fraction"] = scale_cloud_abund(
                    model_param,
                    radtrans.rt_object,
                    pressure,
                    temp,
                    mmw,
                    "equilibrium",
                    abund_in,
                    cloud_item,
                    model_param[f"{cloud_item}_tau"],
                    pressure_grid=radtrans.pressure_grid,
                )

        from petitRADTRANS.chemistry.utils import simplify_species_list

        for cloud_item in radtrans.cloud_species:
            cloud_simple = simplify_species_list([cloud_item])

            cond_temp = get_condensation_curve(
                composition=cloud_simple[0],
                press=pressure,
                metallicity=model_param["metallicity"],
                c_o_ratio=model_param["c_o_ratio"],
                mmw=np.mean(mmw),
            )

            ax.plot(
                cond_temp,
                pressure,
                "--",
                lw=0.8,
                color=next(color_iter, "black"),
                zorder=2,
            )

    if box.attributes["chemistry"] == "free":
        # Remove these parameters otherwise ReadRadtrans.get_model()
        # will assume equilibrium chemistry
        del model_param["metallicity"]
        del model_param["c_o_ratio"]

    if radtrans is not None:
        # Recalculate the best-fit model to update the attributes of radtrans.rt_object
        model_box = radtrans.get_model(model_param)

        contr_1d = np.mean(model_box.contribution, axis=1)
        contr_1d = ax.get_xlim()[0] + 0.5 * (contr_1d / np.amax(contr_1d)) * (
            ax.get_xlim()[1] - ax.get_xlim()[0]
        )

        ax.plot(
            contr_1d,
            1e-6 * radtrans.rt_object.pressures,
            ls="--",
            lw=0.5,
            color="black",
        )

        if extra_axis == "photosphere":
            # Calculate the total optical depth
            # (line and continuum opacities)
            # radtrans.rt_object.calc_opt_depth(10.**model_param['logg'])

            wavelength = radtrans.rt_object.lambda_angstroem * 1e-4  # (um)

            # From Paul: The first axis of total_tau is the coordinate
            # of the cumulative opacity distribution function (ranging
            # from 0 to 1). A correct average is obtained by
            # multiplying the first axis with self.w_gauss, then
            # summing them. This is then the actual wavelength-mean.

            if radtrans.scattering:
                w_gauss = radtrans.rt_object.w_gauss[..., np.newaxis, np.newaxis]

                # From petitRADTRANS: Only use 0 index for species
                # because for lbl or test_ck_shuffle_comp = True
                # everything has been moved into the 0th index
                optical_depth = np.sum(
                    w_gauss * radtrans.rt_object.total_tau[:, :, 0, :], axis=0
                )

            else:
                # TODO Ask Paul if correct
                w_gauss = radtrans.rt_object.w_gauss[
                    ..., np.newaxis, np.newaxis, np.newaxis
                ]
                optical_depth = np.sum(
                    w_gauss * radtrans.rt_object.total_tau[:, :, :, :], axis=0
                )

                # Sum over all species
                optical_depth = np.sum(optical_depth, axis=1)

            ax2 = ax.twiny()

            ax2.tick_params(
                axis="both",
                which="major",
                colors="black",
                labelcolor="black",
                direction="in",
                width=1,
                length=5,
                labelsize=12,
                top=True,
                bottom=False,
                left=True,
                right=True,
            )

            ax2.tick_params(
                axis="both",
                which="minor",
                colors="black",
                labelcolor="black",
                direction="in",
                width=1,
                length=3,
                labelsize=12,
                top=True,
                bottom=False,
                left=True,
                right=True,
            )

            if ylim is None:
                ax2.set_ylim(max_press, 1e-6)
            else:
                ax2.set_ylim(ylim[0], ylim[1])

            ax2.set_yscale("log")

            ax2.set_xlabel(
                "Wavelength (\N{GREEK SMALL LETTER MU}m)", fontsize=13, va="bottom"
            )

            if offset is not None:
                ax2.get_xaxis().set_label_coords(0.5, 1.0 + abs(offset[0]))
            else:
                ax2.get_xaxis().set_label_coords(0.5, 1.06)

            photo_press = np.zeros(wavelength.shape[0])

            for i in range(photo_press.shape[0]):
                # Interpolate the optical depth to
                # the photosphere at tau = 2/3
                press_interp = interp1d(
                    optical_depth[i, :], radtrans.rt_object.pressures
                )
                photo_press[i] = press_interp(2.0 / 3.0) * 1e-6  # cgs to (bar)

            ax2.plot(
                wavelength,
                photo_press,
                lw=0.5,
                color="tab:blue",
                label=r"Photosphere ($\tau$ = 2/3)",
            )

        elif extra_axis == "grains":
            if len(radtrans.cloud_species) > 0:
                ax2 = ax.twiny()

                ax2.tick_params(
                    axis="both",
                    which="major",
                    colors="black",
                    labelcolor="black",
                    direction="in",
                    width=1,
                    length=5,
                    labelsize=12,
                    top=True,
                    bottom=False,
                    left=True,
                    right=True,
                )

                ax2.tick_params(
                    axis="both",
                    which="minor",
                    colors="black",
                    labelcolor="black",
                    direction="in",
                    width=1,
                    length=3,
                    labelsize=12,
                    top=True,
                    bottom=False,
                    left=True,
                    right=True,
                )

                if ylim is None:
                    ax2.set_ylim(max_press, 1e-6)
                else:
                    ax2.set_ylim(ylim[0], ylim[1])

                ax2.set_xscale("log")
                ax2.set_yscale("log")

                ax2.set_xlabel("Average particle radius (µm)", fontsize=13, va="bottom")

                # Recalculate the best-fit model to update the r_g attribute of radtrans.rt_object
                model_box = radtrans.get_model(model_param)

                if offset is not None:
                    ax2.get_xaxis().set_label_coords(0.5, 1.0 + abs(offset[0]))
                else:
                    ax2.get_xaxis().set_label_coords(0.5, 1.06)

            else:
                raise ValueError(
                    "The Radtrans object does not contain any cloud "
                    "species. Please set the argument of 'extra_axis' "
                    "either to 'photosphere' or None."
                )

            color_iter = iter(cloud_colors)

            for cloud_item in radtrans.cloud_species:
                cloud_index = radtrans.rt_object.cloud_species.index(cloud_item)

                label = ""
                for char in cloud_item:
                    if char.isnumeric():
                        label += f"$_{char}$"
                    else:
                        label += char

                if label == "KCL":
                    label = "KCl"

                ax2.plot(
                    # (cm) -> (um)
                    model_box.extra_out["cloud_particles_mean_radii"][:, cloud_index]
                    * 1e4,
                    # (Ba) -> (Bar)
                    radtrans.rt_object.pressures * 1e-6,
                    lw=0.8,
                    color=next(color_iter),
                    label=label,
                )

        if extra_axis is not None:
            ax2.legend(loc="upper right", frameon=False, fontsize=10.0)

    else:
        if extra_axis is not None:
            warnings.warn(
                "The argument of extra_axis is ignored because radtrans does not "
                "contain a ReadRadtrans object."
            )

    if output is None:
        plt.show()
    else:
        plt.savefig(output, bbox_inches="tight")

    print(" [DONE]")

    return fig


@typechecked
def plot_opacities(
    tag: str,
    radtrans: ReadRadtrans,
    offset: Optional[Tuple[float, float]] = None,
    output: Optional[str] = None,
) -> mpl.figure.Figure:
    """
    Function to plot the line and continuum opacity structure of the
    atmosphere by using the median parameters from posterior samples.
    This function is broken and should currently not be used.

    Parameters
    ----------
    tag : str
        Database tag with the posterior samples.
    radtrans : ReadRadtrans
        Instance of :class:`~species.read.read_radtrans.ReadRadtrans`.
        The parameter is not used if the argument is set to ``None``.
    offset : tuple(float, float), None
        Offset of the x- and y-axis label. Default values are used
        if the argument is set to ``None``.
    output : str, None
        Output filename for the plot. The plot is shown in an
        interface window if the argument is set to ``None``.

    Returns
    -------
    matplotlib.figure.Figure
        The ``Figure`` object that can be used for further
        customization of the plot.
    """

    if output is None:
        print("Plotting opacities...", end="", flush=True)
    else:
        print(f"Plotting opacities: {output}...", end="", flush=True)

    from species.data.database import Database

    species_db = Database()
    box = species_db.get_samples(tag)
    model_param = box.prob_sample

    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "dejavuserif"

    fig = plt.figure(figsize=(10.0, 6.0))
    gridsp = mpl.gridspec.GridSpec(2, 5, width_ratios=[4, 0.25, 1.5, 4, 0.25])
    gridsp.update(wspace=0.1, hspace=0.1, left=0, right=1, bottom=0, top=1)

    ax1 = plt.subplot(gridsp[0, 0])
    ax2 = plt.subplot(gridsp[1, 0])
    ax3 = plt.subplot(gridsp[0, 1])
    ax4 = plt.subplot(gridsp[1, 1])

    ax5 = plt.subplot(gridsp[0, 3])
    ax6 = plt.subplot(gridsp[1, 3])
    ax7 = plt.subplot(gridsp[0, 4])
    ax8 = plt.subplot(gridsp[1, 4])

    model_box = radtrans.get_model(model_param)

    # Line opacities

    wavelength, opacity = radtrans.rt_object.__compute_opacities(
        radtrans.rt_object.temperatures
    )

    wavelength *= 1e4  # (um)

    opacity_line = np.zeros(
        (radtrans.rt_object.frequencies.shape[0], radtrans.rt_object.pressures.shape[0])
    )

    for item in radtrans.rt_object.lines_loaded_opacities["opacity_grid"].values():
        opacity_line += item

    # Continuum opacities

    opacity_cont_abs = model_box.extra_out["cloud_absorption_opacities"]
    opacity_cont_scat = model_box.extra_out["continuum_opacities_scattering"]
    opacity_total = model_box.extra_out["opacities"]

    albedo = opacity_cont_scat / opacity_total

    ax1.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
        labelbottom=False,
    )

    ax1.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
        labelbottom=False,
    )

    ax2.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax2.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax3.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax3.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax4.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax4.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax5.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
        labelbottom=False,
    )

    ax5.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
        labelbottom=False,
    )

    ax6.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax6.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax7.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax7.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax8.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax8.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax1.xaxis.set_major_locator(MultipleLocator(1.0))
    ax2.xaxis.set_major_locator(MultipleLocator(1.0))

    ax1.xaxis.set_minor_locator(MultipleLocator(0.2))
    ax2.xaxis.set_minor_locator(MultipleLocator(0.2))

    ax5.xaxis.set_major_locator(MultipleLocator(1.0))
    ax6.xaxis.set_major_locator(MultipleLocator(1.0))

    ax5.xaxis.set_minor_locator(MultipleLocator(0.2))
    ax6.xaxis.set_minor_locator(MultipleLocator(0.2))

    # ax1.yaxis.set_major_locator(LogLocator(base=10.))
    # ax2.yaxis.set_major_locator(LogLocator(base=10.))
    # ax3.yaxis.set_major_locator(LogLocator(base=10.))
    # ax4.yaxis.set_major_locator(LogLocator(base=10.))

    # ax1.yaxis.set_minor_locator(LogLocator(base=1.))
    # ax2.yaxis.set_minor_locator(LogLocator(base=1.))
    # ax3.yaxis.set_minor_locator(LogLocator(base=1.))
    # ax4.yaxis.set_minor_locator(LogLocator(base=1.))

    xx_grid, yy_grid = np.meshgrid(wavelength, 1e-6 * radtrans.rt_object.pressures)

    fig_1 = ax1.pcolormesh(
        xx_grid,
        yy_grid,
        np.transpose(opacity_line),
        cmap="viridis",
        shading="gouraud",
        norm=LogNorm(vmin=1e-6 * np.amax(opacity_line), vmax=np.amax(opacity_line)),
    )

    cb = Colorbar(ax=ax3, mappable=fig_1, orientation="vertical", ticklocation="right")
    cb.ax.set_ylabel("Line opacity (cm$^2$/g)", rotation=270, labelpad=20, fontsize=11)

    fig_2 = ax2.pcolormesh(
        xx_grid,
        yy_grid,
        np.transpose(albedo),
        cmap="viridis",
        shading="gouraud",
        norm=LogNorm(vmin=1e-4 * np.amax(albedo), vmax=np.amax(albedo)),
    )

    cb = Colorbar(ax=ax4, mappable=fig_2, orientation="vertical", ticklocation="right")
    cb.ax.set_ylabel("Single scattering albedo", rotation=270, labelpad=20, fontsize=11)

    fig_3 = ax5.pcolormesh(
        xx_grid,
        yy_grid,
        np.transpose(opacity_cont_abs),
        cmap="viridis",
        shading="gouraud",
        norm=LogNorm(
            vmin=1e-6 * np.amax(opacity_cont_abs), vmax=np.amax(opacity_cont_abs)
        ),
    )

    cb = Colorbar(ax=ax7, mappable=fig_3, orientation="vertical", ticklocation="right")
    cb.ax.set_ylabel(
        "Continuum absorption (cm$^2$/g)", rotation=270, labelpad=20, fontsize=11
    )

    fig_4 = ax6.pcolormesh(
        xx_grid,
        yy_grid,
        np.transpose(opacity_cont_scat),
        cmap="viridis",
        shading="gouraud",
        norm=LogNorm(
            vmin=1e-6 * np.amax(opacity_cont_scat), vmax=np.amax(opacity_cont_scat)
        ),
    )

    cb = Colorbar(ax=ax8, mappable=fig_4, orientation="vertical", ticklocation="right")
    cb.ax.set_ylabel(
        "Continuum scattering (cm$^2$/g)", rotation=270, labelpad=20, fontsize=11
    )

    ax1.set_ylabel("Pressure (bar)", fontsize=13)

    ax2.set_xlabel("Wavelength (\N{GREEK SMALL LETTER MU}m)", fontsize=13)
    ax2.set_ylabel("Pressure (bar)", fontsize=13)

    ax5.set_ylabel("Pressure (bar)", fontsize=13)

    ax6.set_xlabel("Wavelength (\N{GREEK SMALL LETTER MU}m)", fontsize=13)
    ax6.set_ylabel("Pressure (bar)", fontsize=13)

    ax1.set_xlim(wavelength[0], wavelength[-1])
    ax2.set_xlim(wavelength[0], wavelength[-1])

    ax5.set_xlim(wavelength[0], wavelength[-1])
    ax6.set_xlim(wavelength[0], wavelength[-1])

    ax1.set_ylim(
        radtrans.rt_object.pressures[-1] * 1e-6, radtrans.rt_object.pressures[0] * 1e-6
    )
    ax2.set_ylim(
        radtrans.rt_object.pressures[-1] * 1e-6, radtrans.rt_object.pressures[0] * 1e-6
    )

    ax5.set_ylim(
        radtrans.rt_object.pressures[-1] * 1e-6, radtrans.rt_object.pressures[0] * 1e-6
    )
    ax6.set_ylim(
        radtrans.rt_object.pressures[-1] * 1e-6, radtrans.rt_object.pressures[0] * 1e-6
    )

    if offset is not None:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])
        ax1.get_yaxis().set_label_coords(offset[1], 0.5)

        ax2.get_xaxis().set_label_coords(0.5, offset[0])
        ax2.get_yaxis().set_label_coords(offset[1], 0.5)

        ax5.get_xaxis().set_label_coords(0.5, offset[0])
        ax5.get_yaxis().set_label_coords(offset[1], 0.5)

        ax6.get_xaxis().set_label_coords(0.5, offset[0])
        ax6.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax1.get_xaxis().set_label_coords(0.5, -0.1)
        ax1.get_yaxis().set_label_coords(-0.14, 0.5)

        ax2.get_xaxis().set_label_coords(0.5, -0.1)
        ax2.get_yaxis().set_label_coords(-0.14, 0.5)

        ax5.get_xaxis().set_label_coords(0.5, -0.1)
        ax5.get_yaxis().set_label_coords(-0.14, 0.5)

        ax6.get_xaxis().set_label_coords(0.5, -0.1)
        ax6.get_yaxis().set_label_coords(-0.14, 0.5)

    ax1.set_yscale("log")
    ax2.set_yscale("log")
    ax3.set_yscale("log")
    ax4.set_yscale("log")

    ax5.set_yscale("log")
    ax6.set_yscale("log")
    ax7.set_yscale("log")
    ax8.set_yscale("log")

    if output is None:
        plt.show()
    else:
        plt.savefig(output, bbox_inches="tight")

    print(" [DONE]")

    return fig


@typechecked
def plot_clouds(
    tag: str,
    offset: Optional[Tuple[float, float]] = None,
    output: Optional[str] = None,
    radtrans: Optional[ReadRadtrans] = None,
    composition: str = "MgSiO3",
) -> mpl.figure.Figure:
    """
    Function to plot the size distributions for a given cloud
    composition as function as pressure. The size distributions are
    calculated for the median sample by using the ``radius_g`` (as
    function of pressure) and ``sigma_g``.

    Parameters
    ----------
    tag : str
        Database tag with the posterior samples.
    offset : tuple(float, float), None
        Offset of the x- and y-axis label. Default values are
        used if the argument set to ``None``.
    output : str, None
        Output filename for the plot. The plot is shown in an
        interface window if the argument is set to ``None``.
    radtrans : ReadRadtrans, None
        Instance of :class:`~species.read.read_radtrans.ReadRadtrans`.
        The parameter is not used if the argument is set to ``None``.
    composition : str
        Cloud composition (e.g. 'MgSiO3', 'Fe', 'Al2O3', 'Na2S', 'KCl').

    Returns
    -------
    matplotlib.figure.Figure
        The ``Figure`` object that can be used for
        further customization of the plot.
    """

    from species.data.database import Database

    species_db = Database()
    box = species_db.get_samples(tag)
    model_param = box.prob_sample

    if (
        f"{composition}_fraction" not in model_param
        and "log_tau_cloud" not in model_param
        and composition not in model_param
    ):
        raise ValueError(
            f"The mass fraction of the {composition} clouds is "
            "not found. The median sample contains the following "
            f"parameters: {list(model_param.keys())}"
        )

    if output is None:
        print(f"Plotting {composition} clouds...", end="", flush=True)
    else:
        print(f"Plotting {composition} clouds: {output}...", end="", flush=True)

    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "dejavuserif"

    fig = plt.figure(figsize=(4.0, 3.0))
    gridsp = mpl.gridspec.GridSpec(1, 2, width_ratios=[4, 0.25])
    gridsp.update(wspace=0.1, hspace=0.0, left=0, right=1, bottom=0, top=1)

    ax1 = plt.subplot(gridsp[0, 0])
    ax2 = plt.subplot(gridsp[0, 1])

    model_box = radtrans.get_model(model_param)

    cloud_index = radtrans.rt_object.cloud_species.index(composition)
    radius_g = (
        model_box.extra_out["cloud_particles_mean_radii"][:, cloud_index] * 1e4
    )  # (cm) -> (um)
    sigma_g = model_param["sigma_lnorm"]

    r_bins = np.logspace(-3.0, 3.0, 1000)
    radii = (r_bins[1:] + r_bins[:-1]) / 2.0

    dn_dr = np.zeros((radius_g.shape[0], radii.shape[0]))

    for i, item in enumerate(radius_g):
        dn_dr[i,] = lognorm.pdf(radii, s=np.log(sigma_g), loc=0.0, scale=item)

    ax1.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
        labelbottom=True,
    )

    ax1.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
        labelbottom=True,
    )

    ax2.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax2.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    xx_grid, yy_grid = np.meshgrid(radii, 1e-6 * radtrans.rt_object.pressures)

    mesh_fig = ax1.pcolormesh(
        xx_grid,
        yy_grid,
        dn_dr,
        cmap="viridis",
        shading="auto",
        norm=LogNorm(vmin=1e-10 * np.amax(dn_dr), vmax=np.amax(dn_dr)),
    )

    cb = Colorbar(
        ax=ax2, mappable=mesh_fig, orientation="vertical", ticklocation="right"
    )
    cb.ax.set_ylabel("dn/dr", rotation=270, labelpad=20, fontsize=11)

    for item in radtrans.rt_object.pressures * 1e-6:  # (bar)
        ax1.axhline(item, ls="-", lw=0.1, color="white")

    for item in (
        radtrans.rt_object._clouds_loaded_opacities["particles_radii"] * 1e4
    ):  # (um)
        ax1.axvline(item, ls="-", lw=0.1, color="white")

    ax1.text(
        0.05,
        0.05,
        rf"$\sigma_\mathrm{{g}}$ = {sigma_g:.2f}",
        ha="left",
        va="bottom",
        transform=ax1.transAxes,
        color="white",
        fontsize=12.0,
    )

    ax1.set_ylabel("Pressure (bar)", fontsize=13)
    ax1.set_xlabel("Grain radius (µm)", fontsize=13)

    ax1.set_xlim(radii[0], radii[-1])
    ax1.set_ylim(
        radtrans.rt_object.pressures[-1] * 1e-6, radtrans.rt_object.pressures[0] * 1e-6
    )

    if offset is not None:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])
        ax1.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax1.get_xaxis().set_label_coords(0.5, -0.1)
        ax1.get_yaxis().set_label_coords(-0.15, 0.5)

    ax1.set_xscale("log")
    ax1.set_yscale("log")
    ax2.set_yscale("log")

    if output is None:
        plt.show()
    else:
        plt.savefig(output, bbox_inches="tight")

    print(" [DONE]")

    return fig


@typechecked
def plot_abundances(
    tag: str,
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    offset: Optional[Tuple[float, float]] = None,
    output: Optional[str] = None,
    legend: Optional[dict] = None,
    radtrans: Optional[ReadRadtrans] = None,
) -> mpl.figure.Figure:
    """
    Function to plotting the retrieved abundance profiles.

    Parameters
    ----------
    tag : str
        Database tag with the posterior samples.
    xlim : tuple(float, float), None
        Limits of the temperature axis. Default values are used if
        set to ``None``.
    ylim : tuple(float, float), None
        Limits of the pressure axis. Default values are used if set
        to ``None``.
    offset : tuple(float, float), None
        Offset of the x- and y-axis label. Default values are
        used if the argument set to ``None``.
    output : str, None
        Output filename for the plot. The plot is shown in an
        interface window if the argument is set to ``None``.
    legend : dict, None
        Dictionary with legend properties. Default values will
        be used if the argument is set to ``None``.
    radtrans : ReadRadtrans, None
        Instance of :class:`~species.read.read_radtrans.ReadRadtrans`.
        The parameter is not used if the argument is set to ``None``.

    Returns
    -------
    matplotlib.figure.Figure
        The ``Figure`` object that can be used for
        further customization of the plot.
    """

    from species.data.database import Database

    species_db = Database()
    box = species_db.get_samples(tag)
    model_param = box.prob_sample

    if output is None:
        print("Plotting abundances...", end="", flush=True)
    else:
        print(f"Plotting abundances: {output}...", end="", flush=True)

    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "dejavuserif"

    fig = plt.figure(figsize=(4.0, 3.0))
    gridsp = mpl.gridspec.GridSpec(1, 1)
    gridsp.update(wspace=0.1, hspace=0.0, left=0, right=1, bottom=0, top=1)

    ax = plt.subplot(gridsp[0, 0])

    model_box = radtrans.get_model(model_param)

    ax.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
        labelbottom=True,
    )

    ax.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
        labelbottom=True,
    )

    sub_num = str.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉")

    for line_idx, line_item in enumerate(radtrans.rt_object.line_species):
        line_label = line_item.split("_")[0]
        line_label = line_label.translate(sub_num)

        ax.plot(
            radtrans.rt_object.line_abundances[:, line_idx],
            radtrans.rt_object.pressures * 1e-6,
            lw=0.7,
            label=line_label,
        )

    ax.set_xlabel("Abundance", fontsize=13)
    ax.set_ylabel("Pressure (bar)", fontsize=13)

    if xlim is not None:
        ax.set_xlim(1e-10, 1.0)

    if ylim is None:
        ax.set_ylim(
            1e-6 * radtrans.rt_object.pressures[-1],
            1e-6 * radtrans.rt_object.pressures[0],
        )
    else:
        ax.set_ylim(ylim[0], ylim[1])

    ax.set_xscale("log")
    ax.set_yscale("log")

    if offset is not None:
        ax.get_xaxis().set_label_coords(0.5, offset[0])
        ax.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax.get_xaxis().set_label_coords(0.5, -0.1)
        ax.get_yaxis().set_label_coords(-0.15, 0.5)

    if legend is None:
        ax.legend(loc="upper left", fontsize=7.0)
    else:
        ax.legend(**legend)

    if output is None:
        plt.show()
    else:
        plt.savefig(output, bbox_inches="tight")

    print(" [DONE]")

    return fig



from pathlib import Path

import h5py
import numpy as np
import pooch

from typeguard import typechecked

from species.core import constants
from species.util.data_util import extract_tarfile


@typechecked
def add_sonora_bobcat(database: h5py._hl.files.File, input_path: str) -> None:
    """
    Function for adding the isochrone data of `Sonora Bobcat
    <https://zenodo.org/record/5063476>`_ to the database.

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    input_path : str
        Folder where the data is located.

    Returns
    -------
    NoneType
        None
    """

    url = "https://zenodo.org/record/5063476/files/evolution_and_photometery.tar.gz"

    input_file = "evolution_and_photometery.tar.gz"
    data_folder = Path(input_path) / "sonora-bobcat-evolution"
    data_file = data_folder / input_file

    if not data_folder.exists():
        data_folder.mkdir()

    if not data_file.exists():
        print()

        pooch.retrieve(
            url=url,
            known_hash="2198426d1ca0e410fda7b63c3b7f45f3890a8d9f2fcf0a3a1e36e14185283ca5",
            fname=input_file,
            path=data_folder,
            progressbar=True,
        )

    print("\nUnpacking Sonora Bobcat evolution (929 kB)...", end="", flush=True)
    extract_tarfile(str(data_file), str(data_folder))
    print(" [DONE]")

    iso_files = [
        "evo_tables+0.0/nc+0.0_co1.0_age",
        "evo_tables+0.5/nc+0.5_co1.0_age",
        "evo_tables-0.5/nc-0.5_co1.0_age",
    ]

    labels = ["[M/H] = +0.0", "[M/H] = +0.5", "[M/H] = -0.5"]

    for iso_idx, iso_item in enumerate(iso_files):
        iso_file = f"evolution_tables/{iso_item}"
        iso_path = Path(data_folder) / iso_file

        iso_data = []

        with open(str(iso_path), encoding="utf-8") as open_file:
            for line_idx, line_item in enumerate(open_file):
                if line_idx == 0 or " " not in line_item.strip():
                    continue

                # age(Gyr)  M/Msun  log(L/Lsun)  Teff(K)  log(g)  R/Rsun
                param = list(filter(None, line_item.strip().split(" ")))
                param = list(map(float, param))

                param[0] = 1e3 * param[0]  # (Gyr) -> (Myr)
                param[1] = (
                    param[1] * constants.M_SUN / constants.M_JUP
                )  # (Msun) -> (Mjup)
                param[5] = (
                    param[5] * constants.R_SUN / constants.R_JUP
                )  # (Rsun) -> (Rjup)

                iso_data.append(
                    [param[0], param[1], param[2], param[3], param[4], param[5]]
                )

            print(
                f"\nAdding isochrones: Sonora {labels[iso_idx]}...", end="", flush=True
            )

            iso_data = np.array(iso_data)

            metallicity = labels[iso_idx].split(" ")[2]

            dset = database.create_dataset(
                f"isochrones/sonora-bobcat{metallicity}/age", data=iso_data[:, 0]
            )  # (Myr)

            database.create_dataset(
                f"isochrones/sonora-bobcat{metallicity}/mass", data=iso_data[:, 1]
            )  # (Mjup)

            database.create_dataset(
                f"isochrones/sonora-bobcat{metallicity}/log_lum", data=iso_data[:, 2]
            )  # log(L/Lsun)

            database.create_dataset(
                f"isochrones/sonora-bobcat{metallicity}/teff", data=iso_data[:, 3]
            )  # (K)

            database.create_dataset(
                f"isochrones/sonora-bobcat{metallicity}/log_g", data=iso_data[:, 4]
            )  # log(g)

            database.create_dataset(
                f"isochrones/sonora-bobcat{metallicity}/radius", data=iso_data[:, 5]
            )  # (Rjup)

            dset.attrs["model"] = "sonora-bobcat"

            print(" [DONE]")
            print(f"Database tag: sonora{metallicity}")



from pathlib import Path

import h5py
import numpy as np
import pooch

from typeguard import typechecked

from species.core import constants
from species.util.data_util import extract_tarfile


@typechecked
def add_chabrier2023(database: h5py._hl.files.File, input_path: str) -> None:
    """
    Function for adding the AMES-Cond and AMES-Dusty
    isochrone data to the database.

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    input_path : str
        Folder where the data is located.

    Returns
    -------
    NoneType
        None
    """

    url = "https://home.strw.leidenuniv.nl/~stolker/species/chabrier2023_tracks.tgz"

    iso_tag = "ATMO (Chabrier et al. 2023"
    iso_size = "12 MB"

    data_folder = Path(input_path) / "chabrier2023_tracks"

    if not data_folder.exists():
        data_folder.mkdir()

    input_file = url.rsplit("/", maxsplit=1)[-1]
    data_file = Path(input_path) / input_file

    if not data_file.exists():
        print()

        pooch.retrieve(
            url=url,
            known_hash="2f268839107d7084f2512152e4da4be9fb1220793619899806c1f3fc4b3e4034",
            fname=input_file,
            path=input_path,
            progressbar=True,
        )

    print(f"\nUnpacking {iso_tag} isochrones ({iso_size})...", end="", flush=True)
    extract_tarfile(str(data_file), str(data_folder))
    print(" [DONE]")

    iso_files = [
        "ATMO_CEQ",
        "ATMO_NEQ_weak",
        "ATMO_NEQ_strong",
    ]

    labels = [
        "ATMO equilibrium chemistry (Chabrier et al. 2023)",
        "ATMO weak non-equilibrium chemistry (Chabrier et al. 2023)",
        "ATMO strong non-equilibrium chemistry (Chabrier et al. 2023)",
    ]

    db_tags = [
        "atmo-ceq-chabrier2023",
        "atmo-neq-weak-chabrier2023",
        "atmo-neq-strong-chabrier2023",
    ]

    for iso_idx, iso_item in enumerate(iso_files):
        tar_file = str(data_folder / iso_item) + "_neweos.tar.gz"
        print(f"\nUnpacking {iso_item} isochrones...", end="", flush=True)
        extract_tarfile(tar_file, str(data_folder))
        print(" [DONE]")

        iso_path = Path(data_folder) / iso_item / "MKO_WISE_IRAC_vega"

        # Ignore hidden files
        file_list = sorted(iso_path.glob("[!.]*.txt"))

        for file_idx, file_item in enumerate(file_list):
            # Mass (Msun) - Age (Gyr) - Teff (K) - log(L/Lsun) - Radius (Rsun) - log(g)
            if file_idx == 0:
                iso_data = np.loadtxt(str(file_item))

            else:
                iso_load = np.loadtxt(str(file_item))
                iso_data = np.vstack((iso_data, iso_load))

            with open(str(file_item), encoding="utf-8") as open_file:
                parameters = open_file.readline()
                filter_names = parameters.split()[7:]

        iso_data[:, 0] *= constants.M_SUN / constants.M_JUP  # (Msun) -> (Mjup)
        iso_data[:, 1] *= 1e3  # (Gyr) -> (Myr)
        iso_data[:, 4] *= constants.R_SUN / constants.R_JUP  # (Rsun) -> (Rjup)

        print(f"Adding isochrones: {labels[iso_idx]}...", end="", flush=True)

        dtype = h5py.string_dtype(encoding="utf-8", length=None)

        dset = database.create_dataset(
            f"isochrones/{db_tags[iso_idx]}/filters",
            (np.size(filter_names),),
            dtype=dtype,
        )

        dset[...] = filter_names

        database.create_dataset(
            f"isochrones/{db_tags[iso_idx]}/mass", data=iso_data[:, 0]
        )  # (Mjup)
        dset = database.create_dataset(
            f"isochrones/{db_tags[iso_idx]}/age", data=iso_data[:, 1]
        )  # (Myr)
        database.create_dataset(
            f"isochrones/{db_tags[iso_idx]}/teff", data=iso_data[:, 2]
        )  # (K)
        database.create_dataset(
            f"isochrones/{db_tags[iso_idx]}/log_lum", data=iso_data[:, 3]
        )  # log(L/Lsun)
        database.create_dataset(
            f"isochrones/{db_tags[iso_idx]}/radius", data=iso_data[:, 4]
        )  # (Rjup)
        database.create_dataset(
            f"isochrones/{db_tags[iso_idx]}/log_g", data=iso_data[:, 5]
        )  # log(g)

        database.create_dataset(
            f"isochrones/{db_tags[iso_idx]}/magnitudes", data=iso_data[:, 6:]
        )

        dset.attrs["model"] = "atmo"

        print(" [DONE]")
        print(f"Database tag: {db_tags[iso_idx]}")



"""
Module with functions for plotting results from a spectral
analysis that compares data with a library of empirical
spectra or a grid of model spectra.
"""

import configparser
import os
import warnings

from typing import List, Optional, Tuple, Union

import h5py
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from matplotlib.ticker import AutoMinorLocator
from scipy.interpolate import interp1d, RegularGridInterpolator
from typeguard import typechecked

from species.core import constants
from species.read.read_model import ReadModel
from species.read.read_object import ReadObject
from species.util.dust_util import ism_extinction
from species.util.plot_util import (
    create_model_label,
    create_param_format,
    update_labels,
)
from species.util.spec_util import smooth_spectrum


@typechecked
def plot_statistic(
    tag: str,
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    title: Optional[str] = None,
    offset: Optional[Tuple[float, float]] = None,
    figsize: Optional[Tuple[float, float]] = (4.0, 2.5),
    output: Optional[str] = None,
) -> mpl.figure.Figure:
    """
    Function for plotting the goodness-of-fit statistic from a
    comparison with an empirical spectral library with
    :class:`~species.fit.compare_spectra.CompareSpectra.spectral_type`
    that enables a determination of the spectral type

    Parameters
    ----------
    tag : str
        Database tag where the results from the empirical comparison with
        :class:`~species.fit.compare_spectra.CompareSpectra.spectral_type`
        are stored.
    xlim : tuple(float, float)
        Limits of the spectral type axis in numbers (i.e.
        0=M0, 5=M5, 10=L0, etc.).
    ylim : tuple(float, float)
        Limits of the goodness-of-fit axis.
    title : str
        Plot title.
    offset : tuple(float, float)
        Offset for the label of the x- and y-axis.
    figsize : tuple(float, float)
        Figure size.
    output : str, None
        Output filename for the plot. The plot is shown in an
        interface window if the argument is set to ``None``.

    Returns
    -------
    NoneType
        None
    """

    if output is None:
        print("Plotting goodness-of-fit statistic...", end="")
    else:
        print(f"Plotting goodness-of-fit statistic: {output}...", end="")

    if 'SPECIES_CONFIG' in os.environ:
        config_file = os.environ['SPECIES_CONFIG']
    else:
        config_file = os.path.join(os.getcwd(), "species_config.ini")

    config = configparser.ConfigParser()
    config.read(config_file)

    db_path = config["species"]["database"]

    with h5py.File(db_path, "r") as hdf5_file:
        dset = hdf5_file[f"results/empirical/{tag}/names"]

        names = np.array(dset)
        sptypes = np.array(hdf5_file[f"results/empirical/{tag}/sptypes"])
        g_fit = np.array(hdf5_file[f"results/empirical/{tag}/goodness_of_fit"])

    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "dejavuserif"
    plt.rcParams["axes.axisbelow"] = False

    fig = plt.figure(figsize=figsize)
    gridsp = mpl.gridspec.GridSpec(1, 1)
    gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

    ax = plt.subplot(gridsp[0, 0])

    ax.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax.xaxis.set_minor_locator(AutoMinorLocator(5))
    ax.yaxis.set_minor_locator(AutoMinorLocator(5))

    ax.set_xlabel("Spectral type", fontsize=13)
    ax.set_ylabel(r"G$_\mathregular{k}$", fontsize=13)

    if offset is not None:
        ax.get_xaxis().set_label_coords(0.5, offset[0])
        ax.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax.get_xaxis().set_label_coords(0.5, -0.1)
        ax.get_yaxis().set_label_coords(-0.1, 0.5)

    if title is not None:
        ax.set_title(title, y=1.02, fontsize=13)

    ax.set_xticks(np.linspace(0.0, 30.0, 7, endpoint=True))
    ax.set_xticklabels(["M0", "M5", "L0", "L5", "T0", "T5", "Y0"])

    if xlim is None:
        ax.set_xlim(0.0, 30.0)
    else:
        ax.set_xlim(xlim[0], xlim[1])

    if ylim is not None:
        ax.set_ylim(ylim[0], ylim[1])

    sptype_num = np.zeros(names.shape[0])

    for i, item in enumerate(sptypes):
        for j in range(10):
            if not isinstance(item, str):
                item = item.decode("utf-8")

            if item == f"M{j}":
                sptype_num[i] = float(j)

            elif item == f"L{j}":
                sptype_num[i] = float(10 + j)

            elif item == f"T{j}":
                sptype_num[i] = float(20 + j)

    ax.plot(
        sptype_num,
        g_fit,
        "s",
        ms=3.0,
        mew=0.5,
        color="lightgray",
        markeredgecolor="darkgray",
    )

    if output is None:
        plt.show()
    else:
        plt.savefig(output, bbox_inches="tight")

    print(" [DONE]")

    return fig


@typechecked
def plot_empirical_spectra(
    tag: str,
    n_spectra: Optional[int] = None,
    flux_offset: Optional[float] = None,
    label_pos: Optional[Tuple[float, float]] = None,
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    title: Optional[str] = None,
    offset: Optional[Tuple[float, float]] = None,
    figsize: Optional[Tuple[float, float]] = (4.0, 2.5),
    output: Optional[str] = None,
) -> mpl.figure.Figure:
    """
    Function for plotting the results from the empirical
    spectrum comparison.

    Parameters
    ----------
    tag : str
        Database tag where the results from the empirical
        comparison with
        :class:`~species.fit.compare_spectra.CompareSpectra.spectral_type`
        are stored.
    n_spectra : int, None
        The number of spectra with the lowest goodness-of-fit
        statistic that will be plotted in comparison with the data.
        All spectra are selected if the argument is set to ``None``.
    label_pos : tuple(float, float), None
        Position for the name labels. Should be provided as (x, y)
        for the lowest spectrum. The ``flux_offset`` will be applied
        to the remaining spectra. The labels are only plotted if the
        argument of both ``label_pos`` and ``flux_offset`` are not
        ``None``.
    flux_offset : float, None
        Offset to be applied such that the spectra do not overlap. No
        offset is applied if the argument is set to ``None``.
    xlim : tuple(float, float)
        Limits of the spectral type axis.
    ylim : tuple(float, float)
        Limits of the goodness-of-fit axis.
    title : str
        Plot title.
    offset : tuple(float, float)
        Offset for the label of the x- and y-axis.
    figsize : tuple(float, float)
        Figure size.
    output : str, None
        Output filename for the plot. The plot is shown in an
        interface window if the argument is set to ``None``.

    Returns
    -------
    matplotlib.figure.Figure
        The ``Figure`` object that can be used for further
        customization of the plot.
    """

    if output is None:
        print("Plotting empirical spectra comparison...", end="")
    else:
        print(f"Plotting empirical spectra comparison: {output}...", end="")

    if flux_offset is None:
        flux_offset = 0.0

    if 'SPECIES_CONFIG' in os.environ:
        config_file = os.environ['SPECIES_CONFIG']
    else:
        config_file = os.path.join(os.getcwd(), "species_config.ini")

    config = configparser.ConfigParser()
    config.read(config_file)

    db_path = config["species"]["database"]

    with h5py.File(db_path, "r") as hdf5_file:
        dset = hdf5_file[f"results/empirical/{tag}/names"]

        object_name = dset.attrs["object_name"]
        spec_library = dset.attrs["spec_library"]
        n_spec_name = dset.attrs["n_spec_name"]

        spec_name = []
        for i in range(n_spec_name):
            spec_name.append(dset.attrs[f"spec_name{i}"])

        names = np.array(dset)
        flux_scaling = np.array(hdf5_file[f"results/empirical/{tag}/flux_scaling"])
        av_ext = np.array(hdf5_file[f"results/empirical/{tag}/av_ext"])

        rad_vel = np.array(hdf5_file[f"results/empirical/{tag}/rad_vel"])
        rad_vel *= 1e3  # (m s-1)

    if n_spectra is None:
        n_spectra = names.size

    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "dejavuserif"
    plt.rcParams["axes.axisbelow"] = False

    fig = plt.figure(figsize=figsize)
    gridsp = mpl.gridspec.GridSpec(1, 1)
    gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

    ax = plt.subplot(gridsp[0, 0])

    ax.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax.xaxis.set_minor_locator(AutoMinorLocator(5))
    ax.yaxis.set_minor_locator(AutoMinorLocator(5))

    ax.set_xlabel("Wavelength (\N{GREEK SMALL LETTER MU}m)", fontsize=13)

    if flux_offset == 0.0:
        ax.set_ylabel(
            r"$\mathregular{F}_\lambda$"
            + " (W m$^{-2}$ \N{GREEK SMALL LETTER MU}m$^{-1}$)",
            fontsize=11,
        )
    else:
        ax.set_ylabel(
            r"$\mathregular{F}_\lambda$ (W m$^{-2}$"
            + " \N{GREEK SMALL LETTER MU}m$^{-1}$) + offset",
            fontsize=11,
        )

    if xlim is not None:
        ax.set_xlim(xlim[0], xlim[1])

    if ylim is not None:
        ax.set_ylim(ylim[0], ylim[1])

    if offset is not None:
        ax.get_xaxis().set_label_coords(0.5, offset[0])
        ax.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax.get_xaxis().set_label_coords(0.5, -0.1)
        ax.get_yaxis().set_label_coords(-0.1, 0.5)

    if title is not None:
        ax.set_title(title, y=1.02, fontsize=13)

    read_obj = ReadObject(object_name)

    obj_spec = []
    obj_res = []

    for item in spec_name:
        obj_spec.append(read_obj.get_spectrum()[item][0])
        obj_res.append(read_obj.get_spectrum()[item][3])

    if flux_offset == 0.0:
        for spec_item in obj_spec:
            ax.plot(spec_item[:, 0], spec_item[:, 1], "-", lw=0.5, color="black")

    for i in range(n_spectra):
        if isinstance(names[i], str):
            name_item = names[i]
        else:
            name_item = names[i].decode("utf-8")

        with h5py.File(db_path, "r") as hdf5_file:
            dset = hdf5_file[f"spectra/{spec_library}/{name_item}"]
            sptype = dset.attrs["sptype"]
            spectrum = np.asarray(dset)

        if flux_offset != 0.0:
            for spec_item in obj_spec:
                ax.plot(
                    spec_item[:, 0],
                    (n_spectra - i - 1) * flux_offset + spec_item[:, 1],
                    "-",
                    lw=0.5,
                    color="black",
                )

        for j, spec_item in enumerate(obj_spec):
            ism_ext = ism_extinction(av_ext[i], 3.1, spectrum[:, 0])
            ext_scaling = 10.0 ** (-0.4 * ism_ext)

            wavel_shifted = (
                spectrum[:, 0] + spectrum[:, 0] * rad_vel[i] / constants.LIGHT
            )

            flux_smooth = smooth_spectrum(
                wavel_shifted,
                spectrum[:, 1] * ext_scaling,
                spec_res=obj_res[j],
                force_smooth=True,
            )

            interp_spec = interp1d(
                spectrum[:, 0], flux_smooth, fill_value="extrapolate"
            )

            indices = np.where(
                (spec_item[:, 0] > np.amin(spectrum[:, 0]))
                & (spec_item[:, 0] < np.amax(spectrum[:, 0]))
            )[0]

            flux_resample = interp_spec(spec_item[indices, 0])

            ax.plot(
                spec_item[indices, 0],
                (n_spectra - i - 1) * flux_offset + flux_scaling[i][j] * flux_resample,
                color="tomato",
                lw=0.5,
            )

        if label_pos is not None and flux_offset != 0.0:
            label_text = name_item + ", " + sptype

            if av_ext[i] != 0.0:
                label_text += r", A$_\mathregular{V}$ = " + f"{av_ext[i]:.1f}"

            ax.text(
                label_pos[0],
                label_pos[1] + (n_spectra - i - 1) * flux_offset,
                label_text,
                fontsize=8.0,
                ha="left",
            )

    if output is None:
        plt.show()
    else:
        plt.savefig(output, bbox_inches="tight")

    print(" [DONE]")

    return fig


@typechecked
def plot_grid_statistic(
    tag: str,
    upsample: bool = False,
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    title: Optional[str] = None,
    offset: Optional[Tuple[float, float]] = None,
    figsize: Optional[Tuple[float, float]] = (4.0, 2.5),
    output: Optional[str] = None,
    extra_param: Optional[str] = None,
    nlevels_main: int = 20,
    nlevels_extra: int = 10,
    legend: Optional[
        Union[
            str,
            dict,
            Tuple[float, float],
        ]
    ] = "best",
) -> mpl.figure.Figure:
    """
    Function for plotting the results from the comparison with
    a grid of empirical or model spectra.

    Parameters
    ----------
    tag : str
        Database tag where the results from the comparison with
        :class:`~species.fit.compare_spectra.CompareSpectra` are stored.
    upsample : bool
        Upsample the goodness-of-fit grid to a higher resolution
        for a smoother appearance.
    xlim : tuple(float, float), None
        Limits of the x-axis (spectral type or effective temperature).
    ylim : tuple(float, float), None
        Limits of the y-axis.
    title : str, None
        Title that is shown above the plot.
    offset : tuple(float, float), None
        Offset for the label for the x- and y-axis.
    figsize : tuple(float, float), None
        Figure size.
    output : str, None
        Output filename for the plot. The plot is shown in an
        interface window if the argument is set to ``None``.
    extra_param : str, None
        Extra parameter that is shown as contours, overlayed on the
        goodness-of-fit figure. The argument can be set to any of
        the atmospheric parameters that were used for the comparison,
        for example, 'teff', 'logg', 'feh', or 'radius'. Optionally,
        the argument can be set to 'ism_ext' in case the ``av_points``
        parameter of the
        :func:`~species.fit.compare_spectra.CompareSpectra.compare_model`
        method was used. The argument can also be set to 'log_lum' for
        showing the bolometric luminosity. Extra contours are not
        plotted if the argument is set to ``None``.
    nlevels_main : int
        Number of contour levels for the main plot.
    nlevels_extra : int
        Number of contour levels for the optional extra parameter.
    legend : str, tuple, dict, None
        Location of the legend (str or tuple(float, float))
        or a dictionary with the ``**kwargs`` of
        ``matplotlib.pyplot.legend``, for example
        ``{'loc': 'upper left', 'fontsize: 12.}``. The legend is
        not shown when the argument is set to ``None``.

    Returns
    -------
    matplotlib.figure.Figure
        The ``Figure`` object that can be used for further
        customization of the plot.
    """

    if output is None:
        print("Plotting goodness-of-fit of model grid...", end="")
    else:
        print(f"Plotting goodness-of-fit of model grid: {output}...", end="")

    if 'SPECIES_CONFIG' in os.environ:
        config_file = os.environ['SPECIES_CONFIG']
    else:
        config_file = os.path.join(os.getcwd(), "species_config.ini")

    config = configparser.ConfigParser()
    config.read(config_file)

    db_path = config["species"]["database"]

    with h5py.File(db_path, "r") as hdf5_file:
        dset = hdf5_file[f"results/comparison/{tag}/goodness_of_fit"]
        goodness_fit = np.array(dset)

        n_param = dset.attrs["n_param"]
        parallax = dset.attrs["parallax"]

        flux_scaling = np.array(hdf5_file[f"results/comparison/{tag}/flux_scaling"])

        radius = (
            np.sqrt(flux_scaling * (constants.PARSEC / (1e-3 * parallax)) ** 2)
            / constants.R_JUP
        )

        # if 'extra_scaling' in hdf5_file[f'results/comparison/{tag}']:
        #     extra_scaling = np.array(hdf5_file[f'results/comparison/{tag}/extra_scaling'])
        # else:
        #     extra_scaling = None

        model_param = []
        coord_points = []

        for i in range(n_param):
            model_param.append(dset.attrs[f"parameter{i}"])
            coord_points.append(
                np.array(hdf5_file[f"results/comparison/{tag}/coord_points{i}"])
            )

            if dset.attrs[f"parameter{i}"] == "teff":
                teff_idx = i

        object_name = dset.attrs["object_name"]

    read_obj = ReadObject(object_name)

    n_wavel = 0
    for item in read_obj.get_spectrum().values():
        n_wavel += item[0].shape[0]

    # Set the coordinate for the x-axis to Teff
    coord_x = coord_points[0]
    coord_y = None
    param_y = None

    for i, item in enumerate(coord_points[1:]):
        if len(item) > 1:
            # Set the coordinate for the y-axis to the 1st axis after
            # Teff that has a length larger than 1
            coord_y = item
            param_y = model_param[i + 1]
            break

    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "dejavuserif"
    plt.rcParams["axes.axisbelow"] = False

    fig = plt.figure(figsize=figsize)

    if coord_y is None:
        # Create a line plot if there is not a parameter for the y-axis
        gridsp = mpl.gridspec.GridSpec(1, 1)
        gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

        ax = plt.subplot(gridsp[0, 0])

    else:
        # Create a contour plot if there is a second parameter to show
        gridsp = mpl.gridspec.GridSpec(1, 2, width_ratios=[4.0, 0.25])
        gridsp.update(wspace=0.07, hspace=0, left=0, right=1, bottom=0, top=1)

        ax = plt.subplot(gridsp[0, 0])
        ax_cb = plt.subplot(gridsp[0, 1])

    ax.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=11,
        top=True,
        bottom=True,
        left=True,
        right=True,
        pad=5,
    )

    ax.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=11,
        top=True,
        bottom=True,
        left=True,
        right=True,
        pad=5,
    )

    ax.xaxis.set_minor_locator(AutoMinorLocator(5))
    ax.yaxis.set_minor_locator(AutoMinorLocator(5))

    ax.set_xlabel(r"$T_\mathregular{eff}$ (K)", fontsize=13.0)

    if param_y is None:
        ax.set_ylabel(r"$\Delta\log\,G_k$", fontsize=13.0)

    elif param_y == "ism_ext":
        ax.set_ylabel(r"$\mathregular{A}_\mathregular{V}$", fontsize=13.0)

    elif param_y == "logg":
        ax.set_ylabel(r"$\mathregular{log}\,g$", fontsize=13.0)

    elif param_y == "feh":
        ax.set_ylabel("[Fe/H]", fontsize=13.0)

    elif param_y == "c_o_ratio":
        ax.set_ylabel("C/O", fontsize=13.0)

    elif param_y == "fsed":
        ax.set_ylabel(r"$f_\mathregular{sed}$", fontsize=13.0)

    if xlim is not None:
        ax.set_xlim(xlim[0], xlim[1])

    if ylim is not None:
        ax.set_ylim(ylim[0], ylim[1])

    if offset is not None:
        ax.get_xaxis().set_label_coords(0.5, offset[0])
        ax.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax.get_xaxis().set_label_coords(0.5, -0.11)
        ax.get_yaxis().set_label_coords(-0.1, 0.5)

    if title is not None:
        ax.set_title(title, y=1.02, fontsize=14.0)

    # Sum/collapse over parameters with a single value
    for i, item in enumerate(coord_points):
        if len(item) == 1:
            goodness_fit = np.sum(goodness_fit, axis=i)

    # Indices of the best-fit model
    best_index = np.unravel_index(np.nanargmin(goodness_fit), goodness_fit.shape)

    if len(model_param) > 2 or extra_param == "radius":
        if extra_param is None:
            # Select all axes beyond the 1st and 2nd axis
            ax_list = []
            for i in range(len(coord_points) - 2):
                ax_list.append(2 + i)

            # Select minimum G_k along the axes of ax_list
            # This creates a 3D array of goodness_fit
            goodness_fit = np.nanmin(goodness_fit, axis=tuple(ax_list))

        else:
            if extra_param in ["radius", "log_lum"]:
                goodness_full = goodness_fit.copy()

                # Select all axes beyond the 2nd axis
                ax_list = []
                for i in range(len(coord_points) - 2):
                    ax_list.append(2 + i)

                # Select minimum G_k along the axes of ax_list
                # This creates a 2D array from goodness_fit
                goodness_fit = np.nanmin(goodness_fit, axis=tuple(ax_list))

                extra_map = np.zeros(goodness_fit.shape[:2])

                for i in range(goodness_fit.shape[0]):
                    for j in range(goodness_fit.shape[1]):
                        # Get the indices in the goodness_full array
                        # for the values from the collapsed (2D)
                        # goodness_fit array
                        min_idx = np.argwhere(goodness_fit[i, j] == goodness_full)

                        if len(min_idx) == 0:
                            extra_map[i, j] = np.nan
                            continue

                        if len(min_idx) > 1:
                            warnings.warn(
                                f"Found {len(min_idx)} positions in the "
                                "goodness-of-fit grid with the value "
                                f"{goodness_fit[i, j]:.2f}. Using the "
                                "first position in the list but this "
                                "warning is not expected to have occurred."
                            )

                        best_radius = radius[tuple(min_idx[0])]
                        best_teff_idx = min_idx[0][teff_idx]
                        best_teff = coord_points[teff_idx][best_teff_idx]

                        if extra_param == "radius":
                            extra_map[i, j] = best_radius
                        else:
                            extra_map[i, j] = np.log10(
                                4.0
                                * np.pi
                                * (best_radius * constants.R_JUP) ** 2
                                * constants.SIGMA_SB
                                * best_teff**4
                                / constants.L_SUN
                            )

            else:
                extra_idx = model_param.index(extra_param)

                if extra_idx != 2:
                    goodness_fit = np.swapaxes(goodness_fit, extra_idx, 2)

                    coord_points_new = [
                        coord_points[0],
                        coord_points[1],
                        coord_points[extra_idx],
                    ]

                    for coord_idx in range(len(coord_points)):
                        if coord_idx in [0, 1, extra_idx]:
                            continue

                        # Add remaining coordinate points but skip
                        # 1st, 2nd, and extra axis that have already
                        # been added before the for loop
                        coord_points_new.append(coord_points[coord_idx])

                    coord_points = coord_points_new.copy()

                    # Set the extra axis to the 3rd axis after the swap
                    extra_idx = 2

                if len(model_param) > 3:
                    # Select all axes beyond the axis of extra_param
                    ax_list = []
                    for i in range(len(coord_points) - 3):
                        ax_list.append(3 + i)

                    # Select minimum G_k along the axes of ax_list
                    # This creates a 3D array of goodness_fit
                    goodness_fit = np.nanmin(goodness_fit, axis=tuple(ax_list))

                # Indices with the minimum G_k for the tested values
                # of the 3rd axis, that is, the extra_param axis
                # This creates a 2D array with the shape of the
                # 1st and 2nd axis of goodness_fit
                indices = np.argmin(goodness_fit, axis=extra_idx)

                # Select minimum G_k along the 3rd axis
                # This creates a 2D array of goodness_fit
                goodness_fit = np.nanmin(goodness_fit, axis=extra_idx)

                # Create 2D array with the extra map that will
                # be plotted as contours over the main map
                extra_map = np.zeros(goodness_fit.shape)
                for i in range(extra_map.shape[0]):
                    for j in range(extra_map.shape[1]):
                        extra_map[i, j] = coord_points[extra_idx][indices[i, j]]

    # Transpose for plot so make Teff the x axis

    goodness_fit = np.transpose(goodness_fit)

    if extra_param is not None:
        extra_map = np.transpose(extra_map)

    if coord_y is not None:
        if upsample:
            fit_interp = RegularGridInterpolator((coord_y, coord_x), goodness_fit)

            x_new = np.linspace(coord_x[0], coord_x[-1], 50)
            y_new = np.linspace(coord_y[0], coord_y[-1], 50)

            x_grid, y_grid = np.meshgrid(x_new, y_new)

            goodness_fit = fit_interp((y_grid, x_grid))

        else:
            x_grid, y_grid = np.meshgrid(coord_x, coord_y)

    goodness_fit = np.log10(goodness_fit)
    goodness_fit -= np.nanmin(goodness_fit)

    nan_points = np.sum(np.isnan(goodness_fit))
    if nan_points > 0:
        warnings.warn(
            f"Found {nan_points} NaN values in the "
            "goodness-of-fit grid. These points will be set "
            "to zero in the contour map."
        )

        goodness_fit = np.nan_to_num(goodness_fit)

    if coord_y is None:
        ax.plot(
            coord_x,
            goodness_fit[0,],
        )

    else:
        c = ax.contourf(x_grid, y_grid, goodness_fit, levels=nlevels_main)

        cb = mpl.colorbar.Colorbar(
            ax=ax_cb,
            mappable=c,
            orientation="vertical",
            ticklocation="right",
            format="%.1f",
        )

        cb.ax.minorticks_on()

        cb.ax.tick_params(
            which="major",
            width=0.8,
            length=5,
            labelsize=12,
            direction="in",
            color="black",
        )

        cb.ax.tick_params(
            which="minor",
            width=0.8,
            length=3,
            labelsize=12,
            direction="in",
            color="black",
        )

        cb.ax.set_ylabel(
            r"$\Delta\log\,G_k$",
            rotation=270,
            labelpad=22,
            fontsize=13.0,
        )

        if extra_param is not None:
            if upsample:
                extra_interp = RegularGridInterpolator((coord_y, coord_x), extra_map)
                extra_map = extra_interp((y_grid, x_grid))

                cs = ax.contour(
                    x_grid,
                    y_grid,
                    extra_map,
                    levels=nlevels_extra,
                    colors="white",
                    linewidths=0.7,
                )

            else:
                cs = ax.contour(
                    coord_x,
                    coord_y,
                    extra_map,
                    levels=nlevels_extra,
                    colors="white",
                    linewidths=0.7,
                )

            # manual = [(2350, 0.8), (2500, 0.8), (2600, 0.8), (2700, 0.8),
            #           (2800, 0.8), (2950, 0.8), (3100, 0.8), (3300, 0.8)]

            if extra_param == "log_lum":
                ax.clabel(cs, cs.levels, inline=True, fontsize=8, fmt="%1.2f")
            else:
                ax.clabel(cs, cs.levels, inline=True, fontsize=8, fmt="%1.1f")

        # if extra_scaling is not None and len(coord_points[2]) > 1:
        #     ratio = np.transpose(flux_scaling[:, 0, :])/np.transpose(extra_scaling[:, 0, :, 0])
        #
        #     cs = ax.contour(coord_x, coord_y, ratio, levels=nlevels_extra, colors='white',
        #                     linestyles='-', linewidths=0.7)
        #
        #     ax.clabel(cs, cs.levels, inline=True, fontsize=8, fmt='%1.1f')

        ax.plot(
            coord_x[best_index[0]],
            coord_y[best_index[1]],
            marker="X",
            ms=10.0,
            color="#eb4242",
            mfc="#eb4242",
            mec="black",
        )

        # best_param = (coord_x[best_index[0]], coord_y[best_index[1]])
        #
        # par_key, par_unit, par_label = quantity_unit(model_param, object_type='planet')
        #
        # par_text = f'{par_label[0]} = {best_param[0]:.0f} {par_unit[0]}\n' \
        #            f'{par_label[1]} = {best_param[1]:.1f}'
        #
        # ax.annotate(par_text, (best_param[0]+50., best_param[1]), ha='left', va='center',
        #             color='white', fontsize=12.)

    if extra_param is not None:
        extra_label = update_labels([extra_param])[0]
        ax.plot([], [], ls="-", lw=1.2, color="white", label=extra_label)

        if legend is not None:
            if isinstance(legend, (str, tuple)):
                ax.legend(
                    loc=legend, frameon=False, labelcolor="linecolor", fontsize=12.0
                )
            else:
                ax.legend(**legend)

    if output is None:
        plt.show()
    else:
        plt.savefig(output, bbox_inches="tight")

    print(" [DONE]")

    return fig


@typechecked
def plot_model_spectra(
    tag: str,
    n_spectra: Optional[int] = None,
    flux_offset: Optional[float] = None,
    label_pos: Optional[Tuple[float, float]] = None,
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    title: Optional[str] = None,
    offset: Optional[Tuple[float, float]] = None,
    figsize: Optional[Tuple[float, float]] = (4.0, 2.5),
    output: Optional[str] = None,
    leg_param: Optional[List[str]] = None,
) -> mpl.figure.Figure:
    """
    Function for plotting the results from comparing a spectrum
    with a grid of model spectra.

    Parameters
    ----------
    tag : str
        Database tag where the results from the model
        comparison with
        :func:`~species.fit.compare_spectra.CompareSpectra.compare_model`
        are stored.
    n_spectra : int, None
        The number of spectra with the lowest goodness-of-fit
        statistic that will be plotted in comparison with the data.
        All spectra are selected if the argument is set to ``None``.
    label_pos : tuple(float, float), None
        Position for the name labels. Should be provided as (x, y)
        for the lowest spectrum. The ``flux_offset`` will be applied
        to the remaining spectra. The labels are only plotted if the
        argument of both ``label_pos`` and ``flux_offset`` are not
        ``None``.
    flux_offset : float, None
        Offset to be applied such that the spectra do not overlap. No
        offset is applied if the argument is set to ``None``.
    xlim : tuple(float, float), None
        Limits of the wavelength axis.
    ylim : tuple(float, float), None
        Limits of the flux axis.
    title : str
        Plot title.
    offset : tuple(float, float)
        Offset for the label of the x- and y-axis.
    figsize : tuple(float, float)
        Figure size.
    output : str, None
        Output filename for the plot. The plot is shown in an
        interface window if the argument is set to ``None``.
    leg_param : list(str), None
        List with the parameters to include in the legend of the
        model spectra. Apart from atmospheric parameters (e.g.
        'teff', 'logg', 'radius') also parameters such as 'mass'
        and 'log_lum' can be included. The default atmospheric
        parameters are included in the legend if the argument is
        set to ``None``.

    Returns
    -------
    matplotlib.figure.Figure
        The ``Figure`` object that can be used for further
        customization of the plot.
    """

    if output is None:
        print("Plotting model spectra comparison...", end="")
    else:
        print(f"Plotting model spectra comparison: {output}...", end="")

    if flux_offset is None:
        flux_offset = 0.0

    if 'SPECIES_CONFIG' in os.environ:
        config_file = os.environ['SPECIES_CONFIG']
    else:
        config_file = os.path.join(os.getcwd(), "species_config.ini")

    config = configparser.ConfigParser()
    config.read(config_file)

    db_path = config["species"]["database"]

    with h5py.File(db_path, "r") as hdf5_file:
        dset = hdf5_file[f"results/comparison/{tag}/goodness_of_fit"]

        object_name = dset.attrs["object_name"]
        n_spec_name = dset.attrs["n_spec_name"]
        n_param = dset.attrs["n_param"]
        n_scale_spec = dset.attrs["n_scale_spec"]
        parallax = dset.attrs["parallax"]

        if "n_inc_phot" in dset.attrs:
            n_inc_phot = dset.attrs["n_inc_phot"]
        else:
            n_inc_phot = 0

        if "model_name" in dset.attrs:
            model_name = dset.attrs["model_name"]
        else:
            model_name = dset.attrs["model"]

        spec_name = []
        for i in range(n_spec_name):
            spec_name.append(dset.attrs[f"spec_name{i}"])

        inc_phot = []
        for i in range(n_inc_phot):
            inc_phot.append(dset.attrs[f"inc_phot{i}"])

        model_param = []
        coord_points = []
        for i in range(n_param):
            model_param.append(dset.attrs[f"parameter{i}"])
            coord_points.append(
                np.array(hdf5_file[f"results/comparison/{tag}/coord_points{i}"])
            )

        scale_spec = []
        for i in range(n_scale_spec):
            scale_spec.append(dset.attrs[f"scale_spec{i}"])

        goodness_fit = np.array(dset)
        goodness_fit = np.log10(goodness_fit)
        goodness_fit -= np.nanmin(goodness_fit)

        sort_idx = np.unravel_index(
            np.argsort(goodness_fit, axis=None), goodness_fit.shape
        )
        sort_idx = list(sort_idx)

        for i, item in enumerate(sort_idx):
            sort_idx[i] = item[:n_spectra]

        flux_scaling = np.array(hdf5_file[f"results/comparison/{tag}/flux_scaling"])

        if n_scale_spec > 0:
            extra_scaling = np.array(
                hdf5_file[f"results/comparison/{tag}/extra_scaling"]
            )
        else:
            extra_scaling = None

        radius = (
            np.sqrt(flux_scaling * (constants.PARSEC / (1e-3 * parallax)) ** 2)
            / constants.R_JUP
        )

    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "dejavuserif"
    plt.rcParams["axes.axisbelow"] = False

    fig = plt.figure(figsize=figsize)
    gridsp = mpl.gridspec.GridSpec(1, 1)
    gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

    ax = plt.subplot(gridsp[0, 0])

    ax.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax.xaxis.set_minor_locator(AutoMinorLocator(5))
    ax.yaxis.set_minor_locator(AutoMinorLocator(5))

    ax.set_xlabel("Wavelength (\N{GREEK SMALL LETTER MU}m)", fontsize=13)

    if flux_offset == 0.0:
        ax.set_ylabel(
            r"$\mathregular{F}_\lambda$"
            + " (W m$^{-2}$ \N{GREEK SMALL LETTER MU}m$^{-1}$)",
            fontsize=11,
        )
    else:
        ax.set_ylabel(
            r"$\mathregular{F}_\lambda$ (W m$^{-2}$"
            + " \N{GREEK SMALL LETTER MU}m$^{-1}$) + offset",
            fontsize=11,
        )

    if xlim is not None:
        ax.set_xlim(xlim[0], xlim[1])

    if ylim is not None:
        ax.set_ylim(ylim[0], ylim[1])

    if offset is not None:
        ax.get_xaxis().set_label_coords(0.5, offset[0])
        ax.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax.get_xaxis().set_label_coords(0.5, -0.1)
        ax.get_yaxis().set_label_coords(-0.1, 0.5)

    if title is not None:
        ax.set_title(title, y=1.02, fontsize=13)

    read_obj = ReadObject(object_name)
    obj_spec = read_obj.get_spectrum()

    if flux_offset == 0.0:
        for spec_key, spec_item in obj_spec.items():
            if spec_key not in spec_name:
                continue

            ax.plot(spec_item[0][:, 0], spec_item[0][:, 1], "-", lw=0.5, color="black")

    model_reader = ReadModel(model_name)

    for i in range(n_spectra):
        param_select = {"parallax": parallax}
        idx_select = []

        for param_idx, param_item in enumerate(model_param):
            param_select[param_item] = coord_points[param_idx][sort_idx[param_idx][i]]
            idx_select.append(sort_idx[param_idx][i])

        param_select["radius"] = radius[tuple(idx_select)]

        if flux_offset != 0.0:
            for spec_key, spec_item in obj_spec.items():
                if spec_key not in spec_name:
                    continue

                if spec_key in scale_spec:
                    spec_idx = scale_spec.index(spec_key)
                    scaling_idx = np.append(idx_select, spec_idx)
                    data_scaling = extra_scaling[tuple(scaling_idx)]

                else:
                    data_scaling = 1.0

                ax.plot(
                    spec_item[0][:, 0],
                    (n_spectra - i - 1) * flux_offset
                    + data_scaling * spec_item[0][:, 1],
                    "-",
                    lw=0.5,
                    color="black",
                )

        if n_inc_phot > 0 and n_spec_name == 0:
            model_box = model_reader.get_data(
                model_param=param_select,
                spec_res=100.0,
            )

            ax.plot(
                model_box.wavelength,
                (n_spectra - i - 1) * flux_offset + model_box.flux,
                color="tomato",
                lw=0.5,
            )

        for spec_key, spec_item in obj_spec.items():
            if spec_key not in spec_name:
                continue

            model_box = model_reader.get_data(
                model_param=param_select,
                spec_res=spec_item[3],
                wavel_resample=spec_item[0][:, 0],
            )

            ax.plot(
                model_box.wavelength,
                (n_spectra - i - 1) * flux_offset + model_box.flux,
                color="tomato",
                lw=0.5,
            )

        if label_pos is not None and flux_offset != 0.0:
            if leg_param is None:
                leg_param = model_param.copy()
                leg_param.append("radius")

            param_fmt = create_param_format(param_fmt=None)

            label_text = create_model_label(
                model_param=param_select,
                model_name=model_name,
                inc_model_name=False,
                object_type="planet",
                leg_param=leg_param,
                param_fmt=param_fmt,
            )

            label_text = (
                rf"$\Delta\log\,G_k = "
                rf"{goodness_fit[tuple(idx_select)]:.2f}$: {label_text}"
            )

            ax.text(
                label_pos[0],
                label_pos[1] + (n_spectra - i - 1) * flux_offset,
                label_text,
                fontsize=8.0,
                ha="left",
            )

    if output is None:
        plt.show()
    else:
        plt.savefig(output, bbox_inches="tight")

    print(" [DONE]")

    return fig



"""
Utility functions for fit results.
"""

import os
import warnings

from configparser import ConfigParser
from typing import Dict, List, Optional, Union

import h5py
import numpy as np

from spectres.spectral_resampling_numba import spectres_numba
from typeguard import typechecked

from species.core.box import ObjectBox, ResidualsBox, SynphotBox, create_box
from species.phot.syn_phot import SyntheticPhotometry
from species.read.read_calibration import ReadCalibration
from species.read.read_filter import ReadFilter
from species.read.read_model import ReadModel
from species.read.read_planck import ReadPlanck
from species.read.read_radtrans import ReadRadtrans
from species.util.core_util import print_section
from species.util.model_util import binary_to_single, powerlaw_spectrum
from species.util.retrieval_util import convolve_spectrum


@typechecked
def multi_photometry(
    datatype: str,
    spectrum: str,
    filters: List[str],
    parameters: Dict[str, float],
    radtrans: Optional[ReadRadtrans] = None,
    verbose: bool = True,
) -> SynphotBox:
    """
    Function for calculating synthetic photometry for a list of
    filters and a specified atmosphere model and related parameters.
    This function can for example be used for calculating the
    synthetic photometry from a best-fit model spectrum. It returns
    a :class:`~species.core.box_types.SynphotBox` that can be provided
    as input to :func:`~species.plot.plot_spectrum.plot_spectrum`.

    Parameters
    ----------
    datatype : str
        Data type ('model' or 'calibration').
    spectrum : str
        Spectrum name (e.g., 'drift-phoenix', 'bt-settl-cifist',
        planck', 'powerlaw', 'petitradtrans').
    filters : list(str)
        List with the filter names.
    parameters : dict
        Dictionary with the model parameters.
    radtrans : ReadRadtrans, None
        Instance of :class:`~species.read.read_radtrans.ReadRadtrans`.
        Only required with ``spectrum='petitradtrans'`. Make sure that
        the ``wavel_range`` of the ``ReadRadtrans`` instance is
        sufficiently broad to cover all the ``filters``. The argument
        can be set to ``None`` for any other model than petitRADTRANS.
    verbose : bool
        Print output.

    Returns
    -------
    species.core.box.SynphotBox
        Box with synthetic photometry.
    """

    if verbose:
        print_section("Calculate multi-photometry")

        print(f"Data type: {datatype}")
        print(f"Spectrum name: {spectrum}")

        print("\nParameters:")
        for param_key, param_value in parameters.items():
            if isinstance(param_value, float):
                if -0.1 < param_value < 0.1:
                    print(f"   - {param_key} = {param_value:.2e}")
                else:
                    print(f"   - {param_key} = {param_value:.2f}")

    mean_wavel = {}

    for filter_item in filters:
        read_filt = ReadFilter(filter_item)
        mean_wavel[filter_item] = read_filt.mean_wavelength()

    flux = {}

    if datatype == "model":
        if spectrum == "petitradtrans":
            # Calculate the petitRADTRANS spectrum only once
            radtrans_box = radtrans.get_model(parameters)

        for item in filters:
            if spectrum == "petitradtrans":
                # Use an instance of SyntheticPhotometry instead
                # of get_flux from ReadRadtrans in order to not
                # recalculate the spectrum
                syn_phot = SyntheticPhotometry(item)

                flux[item], _ = syn_phot.spectrum_to_flux(
                    radtrans_box.wavelength, radtrans_box.flux
                )

            elif spectrum == "powerlaw":
                synphot = SyntheticPhotometry(item)

                # Set the wavel_range attribute
                synphot.zero_point()

                powerl_box = powerlaw_spectrum(synphot.wavel_range, parameters)
                flux[item] = synphot.spectrum_to_flux(
                    powerl_box.wavelength, powerl_box.flux
                )[0]

            else:
                if spectrum == "planck":
                    readmodel = ReadPlanck(filter_name=item)

                else:
                    readmodel = ReadModel(spectrum, filter_name=item)

                try:
                    if (
                        spectrum != "planck"
                        and "teff_0" in parameters
                        and "teff_1" in parameters
                    ):
                        # Binary system

                        param_0 = binary_to_single(parameters, 0)
                        model_flux_0 = readmodel.get_flux(param_0)[0]

                        param_1 = binary_to_single(parameters, 1)
                        model_flux_1 = readmodel.get_flux(param_1)[0]

                        # Weighted flux of two spectra for atmospheric asymmetries
                        # Or simply the same in case of an actual binary system

                        if "spec_weight" in parameters:
                            flux[item] = (
                                parameters["spec_weight"] * model_flux_0
                                + (1.0 - parameters["spec_weight"]) * model_flux_1
                            )

                        else:
                            flux[item] = model_flux_0 + model_flux_1

                    else:
                        # Single object

                        flux[item] = readmodel.get_flux(parameters)[0]

                except IndexError:
                    flux[item] = np.nan

                    warnings.warn(
                        f"The wavelength range of the {item} filter does not "
                        f"match with the wavelength range of {spectrum}. The "
                        f"flux is set to NaN."
                    )

    elif datatype == "calibration":
        for item in filters:
            readcalib = ReadCalibration(spectrum, filter_name=item)
            flux[item] = readcalib.get_flux(parameters)[0]

    app_mag = {}
    abs_mag = {}

    for phot_key, phot_value in flux.items():
        syn_phot = SyntheticPhotometry(phot_key)
        if "parallax" in parameters:
            app_mag[phot_key], abs_mag[phot_key] = syn_phot.flux_to_magnitude(
                flux=phot_value, error=None, parallax=(parameters["parallax"], None)
            )

        elif "distance" in parameters:
            app_mag[phot_key], abs_mag[phot_key] = syn_phot.flux_to_magnitude(
                flux=phot_value, error=None, distance=(parameters["distance"], None)
            )

        else:
            app_mag[phot_key], abs_mag[phot_key] = syn_phot.flux_to_magnitude(
                flux=phot_value, error=None
            )

    if verbose:
        print("\nMagnitudes:")
        for phot_key, phot_value in app_mag.items():
            if phot_value[1] is None:
                print(f"   - {phot_key} = {phot_value[0]:.2f}")
            else:
                print(f"   - {phot_key} = {phot_value[0]:.2f} +/- {phot_value[1]:.2f}")

        print("\nFluxes (W m-2 um-1):")
        for phot_key, phot_value in flux.items():
            print(f"   - {phot_key} = {phot_value:.2e}")

    return create_box(
        "synphot",
        name="synphot",
        flux=flux,
        wavelength=mean_wavel,
        app_mag=app_mag,
        abs_mag=abs_mag,
    )


@typechecked
def get_residuals(
    parameters: Dict[str, float],
    objectbox: ObjectBox,
    tag: Optional[str] = None,
    inc_phot: Union[bool, List[str]] = True,
    inc_spec: Union[bool, List[str]] = True,
    radtrans: Optional[ReadRadtrans] = None,
    model_name: Optional[str] = None,
    datatype: Optional[str] = None,
    spectrum: Optional[str] = None,
) -> ResidualsBox:
    """
    Function for calculating the residuals from fitting model or
    calibration spectra to a set of spectra and/or photometry.
    This function also calculates the (reduced) :math:`\\chi^2`,
    taking into account the covariances in case these have been
    provided with a spectrum.

    Parameters
    ----------
    parameters : dict
        Parameters and values for the spectrum
    objectbox : ObjectBox
        Box with the photometry and/or spectra of an object. A scaling
        and/or error inflation of the spectra should be applied with
        :func:`~species.util.box_util.update_objectbox` beforehand.
    tag : str, None
        Database tag with the sampling results. By setting the
        argument, the reduced :math:`\\chi^2` will be calculated
        and the ``model_name`` does not need to be set, since the
        model from the fit is automatically selected. Setting the
        argument to ``None`` requires setting the ``model_name``,
        which can be used for calculating the residuals for any
        set of model parameters, not associated with a fit.
    inc_phot : bool, list(str)
        Include photometric data in the fit. If a boolean, either all
        (``True``) or none (``False``) of the data are selected. If a
        list, a subset of filter names (as stored in the database) can
        be provided.
    inc_spec : bool, list(str)
        Include spectroscopic data in the fit. If a boolean, either all
        (``True``) or none (``False``) of the data are selected. If a
        list, a subset of spectrum names (as stored in the database
        with :func:`~species.data.database.Database.add_object`) can be
        provided.
    radtrans : ReadRadtrans, None
        Instance of :class:`~species.read.read_radtrans.ReadRadtrans`.
        Only required with ``spectrum='petitradtrans'``. Make sure that
        the ``wavel_range`` of the ``ReadRadtrans`` instance is
        sufficiently broad to cover all the photometric and
        spectroscopic data of ``inc_phot`` and ``inc_spec``. Not used
        if the argument is set to ``None``.
    model_name : str, None
        The atmospheric model that will be used for calculating the
        residuals. The argument is mandatory when ``tag=None``, and
        optional otherwise. If the argument of ``tag`` is not ``None``
        then the model is automatically selected, but this can be
        overruled by setting the argument of ``model_name``.

    Returns
    -------
    species.core.box.ResidualsBox
        Box with the residuals.
    """

    # Check deprecated parameters

    if datatype is not None:
        warnings.warn(
            "The 'datatype' parameter is no longer "
            "used by the 'get_residuals()' function. "
            "Instead, the 'tag' parameter should be set, "
            "which points to the sampling results as "
            "stored in the database.",
            DeprecationWarning,
        )

    if spectrum is not None:
        warnings.warn(
            "The 'spectrum' parameter is no longer "
            "used by the 'get_residuals()' function. "
            "Instead, the 'tag' parameter should be set, "
            "which points to the sampling results as "
            "stored in the database.",
            DeprecationWarning,
        )

    # Read sampling results

    print_section("Calculate residuals")

    if "SPECIES_CONFIG" in os.environ:
        config_file = os.environ["SPECIES_CONFIG"]
    else:
        config_file = os.path.join(os.getcwd(), "species_config.ini")

    config = ConfigParser()
    config.read(config_file)

    database_path = config["species"]["database"]

    print(f"Database tag: {tag}")

    if tag is not None:
        with h5py.File(database_path, "r") as hdf5_file:
            if f"results/fit/{tag}/samples" in hdf5_file:
                # Samples from FitModel or AtmosphericRetrieval
                dset = hdf5_file[f"results/fit/{tag}/samples"]

                if "model_name" in dset.attrs:
                    if dset.attrs["model_name"] == "petitradtrans":
                        results_type = "AtmosphericRetrieval"
                    else:
                        results_type = "FitModel"

                else:
                    results_type = "FitModel"

                print(f"Results type: {results_type}")

            elif f"results/comparison/{tag}/goodness_of_fit" in hdf5_file:
                # Samples from CompareSpectra
                results_type = "CompareSpectra"
                dset = hdf5_file[f"results/comparison/{tag}/goodness_of_fit"]
                print("Selected results: CompareSpectra")

            else:
                raise ValueError(
                    f"The '{tag}' tag is not found in the database. "
                    "Please specify modeling results from either "
                    "FitModel, AtmosphericRetrieval, or CompareSpectra."
                )

            if model_name is None:
                if "model_name" in dset.attrs:
                    model_name = dset.attrs["model_name"]

                elif "spectrum" in dset.attrs:
                    model_name = dset.attrs["spectrum"]

                elif "model" in dset.attrs:
                    model_name = dset.attrs["model"]

                else:
                    raise ValueError(
                        "The attribute with the model name is "
                        f"not found in the results of '{tag}'."
                    )

            else:
                warnings.warn(
                    "The argument of 'model_name' is set, so the "
                    "model name that is stored as attribute of "
                    "the provided database 'tag' will be ignored."
                )

            print(f"Model: {model_name}")

            if "binary" in dset.attrs:
                binary = dset.attrs["binary"]
                print(f"Binary: {binary}")

            n_param = dset.attrs["n_param"]

            if results_type in ["FitModel", "AtmosphericRetrieval"]:
                if "n_fixed" in dset.attrs:
                    n_fixed = dset.attrs["n_fixed"]

                else:
                    n_fixed = 0

                    warnings.warn(
                        "The 'fixed_param' group is not found in "
                        f"the results of {tag}. Probably the "
                        "results were obtained with an older "
                        "version of the package. The number "
                        "of fixed parameters will be set to "
                        "zero. The reduced chi^2 may not be "
                        "accurate in case parameters were "
                        "fixed when running FitModel."
                    )

            else:
                # TODO not yet implemented for CompareSpectra
                warnings.warn(
                    "Not yet implemented for CompareSpectra. Please "
                    "open an issue on the Github page if needed. The "
                    "number of fixed parameters will be set to zero."
                )
                n_fixed = 0

            print("\nModel parameters:")
            for param_idx in range(n_param):
                param_item = dset.attrs[f"parameter{param_idx}"]
                print(f"   - {param_item}")

            if n_fixed == 0:
                print("\nFixed parameters: none")

            else:
                dset = hdf5_file[f"results/fit/{tag}/fixed_param"]

                print("\nFixed parameters:")
                for param_item in hdf5_file[f"results/fit/{tag}/fixed_param"]:
                    print(f"   - {param_item}")

    print(f"\nInclude photometry: {inc_phot}")
    print(f"Include spectra: {inc_spec}")

    res_phot = None
    res_spec = None

    chi2_stat = 0
    n_dof = 0

    # Photometry residuals

    if inc_phot and objectbox.filters is not None:
        if isinstance(inc_phot, bool) and inc_phot:
            inc_phot = objectbox.filters

        model_phot = multi_photometry(
            datatype="model",
            spectrum=model_name,
            filters=inc_phot,
            parameters=parameters,
            radtrans=radtrans,
            verbose=False,
        )

        res_phot = {}

        for phot_item in inc_phot:
            transmission = ReadFilter(phot_item)
            res_phot[phot_item] = np.zeros(objectbox.flux[phot_item].shape)

            if objectbox.flux[phot_item].ndim == 1:
                res_phot[phot_item][0] = transmission.mean_wavelength()

                res_phot[phot_item][1] = (
                    objectbox.flux[phot_item][0] - model_phot.flux[phot_item]
                ) / objectbox.flux[phot_item][1]

                chi2_stat += res_phot[phot_item][1] ** 2
                n_dof += 1

            elif objectbox.flux[phot_item].ndim == 2:
                for j in range(objectbox.flux[phot_item].shape[1]):
                    res_phot[phot_item][0, j] = transmission.mean_wavelength()

                    res_phot[phot_item][1, j] = (
                        objectbox.flux[phot_item][0, j] - model_phot.flux[phot_item]
                    ) / objectbox.flux[phot_item][1, j]

                    chi2_stat += res_phot[phot_item][1, j] ** 2
                    n_dof += 1

    # Spectra residuals

    if inc_spec and objectbox.spectrum is not None:
        res_spec = {}

        if model_name == "petitradtrans":
            # Calculate the petitRADTRANS spectrum only once
            # Smoothing and resampling not with get_model

            model_box = radtrans.get_model(parameters)

        for spec_key in objectbox.spectrum:
            if isinstance(inc_spec, bool) or spec_key in inc_spec:
                wavel_range = (
                    0.9 * objectbox.spectrum[spec_key][0][0, 0],
                    1.1 * objectbox.spectrum[spec_key][0][-1, 0],
                )

                wl_new = objectbox.spectrum[spec_key][0][:, 0]
                spec_res = objectbox.spectrum[spec_key][3]

                if model_name == "planck":
                    # Resampling to the new wavelength points
                    # is done by the get_spectrum method
                    readmodel = ReadPlanck(wavel_range=wavel_range)

                    model_box = readmodel.get_spectrum(
                        model_param=parameters,
                        spec_res=1000.0,
                        wavel_resample=wl_new,
                    )

                    flux_new = model_box.flux

                elif model_name == "petitradtrans":
                    # Smoothing to the instrument resolution

                    flux_smooth = convolve_spectrum(
                        model_box.wavelength, model_box.flux, spec_res
                    )

                    # Resampling to the new wavelength points

                    flux_new = spectres_numba(
                        wl_new,
                        model_box.wavelength,
                        flux_smooth,
                        spec_errs=None,
                        fill=np.nan,
                        verbose=True,
                    )

                else:
                    # Resampling to the new wavelength points
                    # is done by the get_model method

                    readmodel = ReadModel(model_name, wavel_range=wavel_range)

                    if "teff_0" in parameters and "teff_1" in parameters:
                        # Binary system

                        param_0 = binary_to_single(parameters, 0)

                        model_box_0 = readmodel.get_model(
                            param_0,
                            spec_res=spec_res,
                            wavel_resample=wl_new,
                        )

                        param_1 = binary_to_single(parameters, 1)

                        model_box_1 = readmodel.get_model(
                            param_1,
                            spec_res=spec_res,
                            wavel_resample=wl_new,
                        )

                        # Weighted flux of two spectra for atmospheric asymmetries
                        # Or simply the same in case of an actual binary system

                        if "spec_weight" in parameters:
                            flux_comb = (
                                parameters["spec_weight"] * model_box_0.flux
                                + (1.0 - parameters["spec_weight"]) * model_box_1.flux
                            )

                        else:
                            flux_comb = model_box_0.flux + model_box_1.flux

                        model_box = create_box(
                            boxtype="model",
                            model=model_name,
                            wavelength=wl_new,
                            flux=flux_comb,
                            parameters=parameters,
                            quantity="flux",
                        )

                    else:
                        # Single object

                        model_box = readmodel.get_model(
                            parameters,
                            spec_res=spec_res,
                            wavel_resample=wl_new,
                        )

                    flux_new = model_box.flux

                data_spec = objectbox.spectrum[spec_key][0]
                diff_spec = data_spec[:, 1] - flux_new
                res_tmp = diff_spec / data_spec[:, 2]

                res_spec[spec_key] = np.column_stack([wl_new, res_tmp])
                count_nan = np.sum(np.isnan(res_spec[spec_key][:, 1]))

                if objectbox.spectrum[spec_key][2] is None:
                    if (
                        f"corr_len_{spec_key}" in parameters
                        and f"corr_amp_{spec_key}" in parameters
                    ):
                        # Covariance model (Wang et al. 2020)

                        wavel = objectbox.spectrum[spec_key][0][:, 0]  # (um)
                        wavel_j, wavel_i = np.meshgrid(wavel, wavel)

                        error = objectbox.spectrum[spec_key][0][:, 1]  # (W m-2 um-1)
                        error_j, error_i = np.meshgrid(error, error)

                        corr_len = 10.0 ** parameters[f"corr_len_{spec_key}"]  # (um)
                        corr_amp = parameters[f"corr_amp_{spec_key}"]

                        cov_matrix = (
                            corr_amp**2
                            * error_i
                            * error_j
                            * np.exp(-((wavel_i - wavel_j) ** 2) / (2.0 * corr_len**2))
                            + (1.0 - corr_amp**2) * np.eye(wavel.shape[0]) * error_i**2
                        )

                        chi2_stat += diff_spec @ np.linalg.inv(cov_matrix) @ diff_spec

                    else:
                        chi2_stat += np.nansum(res_spec[spec_key][:, 1] ** 2)

                else:
                    # Use objectbox.spectrum[spec_key][2] which contains
                    # the inverse of the covariance matrix of a spectrum

                    if data_spec.shape[0] != objectbox.spectrum[spec_key][2].shape[0]:
                        raise ValueError(
                            f"The '{spec_key}' spectrum has {data_spec.shape[0]} "
                            "wavelengths but the covariance matrix has been "
                            f"specified for {objectbox.spectrum[spec_key][2].shape[0]} "
                            "wavelengths. Could it be that the spectrum has been "
                            f"resampled? Please make sure that '{spec_key}' spectrum has "
                            "the same number of wavelengths as its covariance matrix."
                        )

                    chi2_stat += diff_spec @ objectbox.spectrum[spec_key][2] @ diff_spec

                n_dof += res_spec[spec_key].shape[0] - count_nan

    print("\nResiduals (sigma):")

    if res_phot is not None:
        for phot_item in inc_phot:
            if res_phot[phot_item].ndim == 1:
                print(f"   - {phot_item} = {res_phot[phot_item][1]:.2f}")

            elif res_phot[phot_item].ndim == 2:
                for j in range(res_phot[phot_item].shape[1]):
                    print(f"   - {phot_item} = {res_phot[phot_item][1, j]:.2f}")

    if res_spec is not None:
        for spec_key in objectbox.spectrum:
            if isinstance(inc_spec, bool) or spec_key in inc_spec:
                print(
                    f"   - {spec_key}: min = {np.nanmin(res_spec[spec_key]):.2f}, "
                    f"max = {np.nanmax(res_spec[spec_key]):.2f}"
                )

    if tag is not None:
        print(f"\nNumber of data points = {n_dof}")
        print(f"Number of model parameters = {n_param}")
        print(f"Number of fixed parameters = {n_fixed}")

        n_dof -= n_param - n_fixed

        print(f"Number of degrees of freedom = {n_dof}")

        chi2_red = chi2_stat / n_dof

    else:
        n_dof = None
        chi2_red = None

    print(f"\nchi2 = {chi2_stat:.2f}")

    if tag is not None:
        print(f"reduced chi2 = {chi2_red:.2f}")

    return create_box(
        boxtype="residuals",
        name=objectbox.name,
        photometry=res_phot,
        spectrum=res_spec,
        chi2=chi2_stat,
        n_dof=n_dof,
        chi2_red=chi2_red,
    )



"""
Module for the photometric data and parallaxes from the
Database of Ultracool Parallaxes.
"""

from pathlib import Path

import h5py
import numpy as np
import pooch

from astropy.io import fits
from typeguard import typechecked

from species.util.data_util import update_sptype


@typechecked
def add_vlm_plx(input_path: str, database: h5py._hl.files.File) -> None:
    """
    Function for adding the Database of Ultracool Parallaxes to the
    database. The FITS file with data was originally downloaded from
    http://www.as.utexas.edu/~tdupuy/plx/Database_of_Ultracool_Parallaxes_files/vlm-plx-all.fits
    but that website has been taken offline, probably because of the
    new table at http://bit.ly/UltracoolSheet.

    Parameters
    ----------
    input_path : str
        Data folder.
    database : h5py._hl.files.File
        The HDF5 database that has been opened.

    Returns
    -------
    NoneType
        None
    """

    input_file = "vlm-plx-all.fits"
    data_file = Path(input_path) / "vlm-plx-all.fits"
    url = "https://home.strw.leidenuniv.nl/~stolker/species/vlm-plx-all.fits"

    if not data_file.exists():
        print()

        pooch.retrieve(
            url=url,
            known_hash="d31bb3162d7de890c09ebf9f0497d51159889b5f5e7c4da1ddf01f24d0c2b36f",
            fname=input_file,
            path=input_path,
            progressbar=True,
        )

    database.create_group("photometry/vlm-plx")

    with fits.open(data_file) as hdu_list:
        phot_data = hdu_list[1].data

    parallax = phot_data["PLX"]  # (mas)
    parallax_error = phot_data["EPLX"]  # (mas)

    name = phot_data["NAME"]
    name = np.core.defchararray.strip(name)

    sptype = phot_data["OSPTSTR"]
    sptype = np.core.defchararray.strip(sptype)

    sptype_nir = phot_data["ISPTSTR"]
    sptype_nir = np.core.defchararray.strip(sptype_nir)

    for i, item in enumerate(sptype):
        if item == "null":
            sptype[i] = sptype_nir[i]

    flag = phot_data["FLAG"]
    flag = np.core.defchararray.strip(flag)

    sptype = update_sptype(sptype)

    dtype = h5py.special_dtype(vlen=str)

    dset = database.create_dataset(
        "photometry/vlm-plx/name", (np.size(name),), dtype=dtype
    )
    dset[...] = name

    dset = database.create_dataset(
        "photometry/vlm-plx/sptype", (np.size(sptype),), dtype=dtype
    )
    dset[...] = sptype

    dset = database.create_dataset(
        "photometry/vlm-plx/flag", (np.size(flag),), dtype=dtype
    )
    dset[...] = flag

    database.create_dataset("photometry/vlm-plx/ra", data=phot_data["RA"])  # (deg)
    database.create_dataset("photometry/vlm-plx/dec", data=phot_data["DEC"])  # (deg)
    database.create_dataset("photometry/vlm-plx/parallax", data=parallax)
    database.create_dataset("photometry/vlm-plx/parallax_error", data=parallax_error)
    database.create_dataset("photometry/vlm-plx/Keck/NIRC.Y", data=phot_data["YMAG"])
    database.create_dataset("photometry/vlm-plx/MKO/NSFCam.J", data=phot_data["JMAG"])
    database.create_dataset("photometry/vlm-plx/MKO/NSFCam.H", data=phot_data["HMAG"])
    database.create_dataset("photometry/vlm-plx/MKO/NSFCam.K", data=phot_data["KMAG"])
    database.create_dataset("photometry/vlm-plx/MKO/NSFCam.Lp", data=phot_data["LMAG"])
    database.create_dataset("photometry/vlm-plx/MKO/NSFCam.Mp", data=phot_data["MMAG"])
    database.create_dataset("photometry/vlm-plx/2MASS/2MASS.J", data=phot_data["J2MAG"])
    database.create_dataset("photometry/vlm-plx/2MASS/2MASS.H", data=phot_data["H2MAG"])
    database.create_dataset(
        "photometry/vlm-plx/2MASS/2MASS.Ks", data=phot_data["K2MAG"]
    )

    database.close()



import os
import shutil

import pytest
import numpy as np

from species import SpeciesInit
from species.data.database import Database
from species.read.read_object import ReadObject
from species.util import test_util


class TestObject:
    def setup_class(self):
        self.limit = 1e-8
        self.test_path = os.path.dirname(__file__) + "/"

    def teardown_class(self):
        os.remove("species_database.hdf5")
        os.remove("species_config.ini")
        shutil.rmtree("data/")

    def test_species_init(self):
        test_util.create_config("./")
        SpeciesInit()

    def test_read_object(self):
        database = Database()

        with pytest.warns(UserWarning):
            database.add_companion(name="beta Pic b")

        read_object = ReadObject("beta Pic b")
        assert read_object.object_name == "beta Pic b"

        with pytest.raises(ValueError) as error:
            ReadObject("wrong name")

        assert (
            str(error.value)
            == "The object 'wrong name' is not present in the database."
        )

    def test_get_photometry(self):
        read_object = ReadObject("beta Pic b")
        photometry = read_object.get_photometry("Paranal/NACO.Lp")

        assert isinstance(photometry, np.ndarray)

        assert photometry[0] == pytest.approx(11.3, rel=self.limit, abs=0.0)
        assert photometry[1] == pytest.approx(0.06, rel=self.limit, abs=0.0)
        assert photometry[2] == pytest.approx(
            1.5566122192562612e-15, rel=self.limit, abs=0.0
        )
        assert photometry[3] == pytest.approx(
            8.606536033479756e-17, rel=self.limit, abs=0.0
        )

    def test_get_parallax(self):
        read_object = ReadObject("beta Pic b")
        parallax = read_object.get_parallax()

        assert parallax == (50.9307, 0.1482)

    def test_get_distance(self):
        read_object = ReadObject("beta Pic b")
        distance = read_object.get_distance()

        assert distance == (19.63452298908124, 0.05713373162362245)

    def test_get_absmag(self):
        read_object = ReadObject("beta Pic b")
        abs_mag = read_object.get_absmag("Paranal/NACO.Lp")

        assert abs_mag[0] == pytest.approx(9.834898226163453, rel=self.limit, abs=0.0)
        assert abs_mag[1] == pytest.approx(0.06033179718686261, rel=self.limit, abs=0.0)



from pathlib import Path

import h5py
import numpy as np
import pooch

from typeguard import typechecked

from species.core import constants


@typechecked
def add_parsec(database: h5py._hl.files.File, input_path: str) -> None:
    """
    Function for adding the PARSEC v2.0 isochrone data to the database.

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    input_path : str
        Folder where the data is located.

    Returns
    -------
    NoneType
        None
    """

    iso_tag = "parsec"

    url = "https://home.strw.leidenuniv.nl/~stolker/species/parsec_evolution.dat"

    input_file = url.rsplit("/", maxsplit=1)[-1]
    data_file = Path(input_path) / input_file

    if not data_file.exists():
        print()

        pooch.retrieve(
            url=url,
            known_hash="c41f60460ac0bf89390b16645238f9cc692316ce158543634763e6c928115b6e",
            fname=input_file,
            path=input_path,
            progressbar=True,
        )

    iso_data = np.loadtxt(
        data_file,
        comments="#",
        delimiter=None,
        usecols=[2, 5, 6, 7, 8, 32, 33],
        unpack=False,
    )

    idx_bad = iso_data[:, 2] == -9.999
    iso_data = iso_data[~idx_bad, :]

    log_age, mass, log_lum, log_teff, log_g, radius_pol, radius_eq = iso_data.T

    age = 1e-6 * 10.0**log_age  # (Myr)
    mass *= constants.M_SUN / constants.M_JUP  # (Msun) -> (Mjup)
    teff = 10.0**log_teff  # (K)
    radius = (radius_pol + radius_eq) / 2.0  # (Rsun)
    radius *= constants.R_SUN / constants.R_JUP  # (Rjup)

    print("Adding isochrones: PARSEC v2.0...", end="", flush=True)

    database.create_dataset(f"isochrones/{iso_tag}/mass", data=mass)  # (Mjup)
    dset = database.create_dataset(f"isochrones/{iso_tag}/age", data=age)  # (Myr)
    database.create_dataset(f"isochrones/{iso_tag}/teff", data=teff)  # (K)
    database.create_dataset(
        f"isochrones/{iso_tag}/log_lum", data=log_lum
    )  # log(L/Lsun)
    database.create_dataset(f"isochrones/{iso_tag}/radius", data=radius)  # (Rjup)
    database.create_dataset(f"isochrones/{iso_tag}/log_g", data=log_g)  # log(g)

    dset.attrs["model"] = iso_tag

    print(" [DONE]")
    print(f"Database tag: {iso_tag}")



import os
import shutil

import pytest
import numpy as np

from species import SpeciesInit
from species.read.read_filter import ReadFilter
from species.util import test_util


class TestFilter:
    def setup_class(self):
        self.limit = 1e-8
        self.test_path = os.path.dirname(__file__) + "/"

    def teardown_class(self):
        os.remove("species_database.hdf5")
        os.remove("species_config.ini")
        shutil.rmtree("data/")

    def test_species_init(self):
        test_util.create_config("./")
        SpeciesInit()

    def test_read_filter(self):
        read_filter = ReadFilter("MKO/NSFCam.H")

        assert read_filter.filter_name == "MKO/NSFCam.H"

    def test_get_filter(self):
        read_filter = ReadFilter("MKO/NSFCam.H")
        filter_profile = read_filter.get_filter()

        assert filter_profile.shape == (970, 2)
        assert np.sum(filter_profile) == pytest.approx(2089.2432, rel=1e-6, abs=0.0)

    def test_detector_type(self):
        read_filter = ReadFilter("MKO/NSFCam.H")
        det_type = read_filter.detector_type()

        assert det_type == "energy"

    def test_get_filter_photon_counter(self):
        read_filter = ReadFilter("Keck/NIRC2.J")
        filter_profile = read_filter.get_filter()

        assert filter_profile.shape == (1054, 2)
        assert np.sum(filter_profile) == pytest.approx(2038.0486, rel=1e-6, abs=0.0)

    def test_detector_type_photon(self):
        read_filter = ReadFilter("Keck/NIRC2.J")
        det_type = read_filter.detector_type()

        assert det_type == "photon"

    def test_interpolate_filter(self):
        read_filter = ReadFilter("MKO/NSFCam.H")
        interp_filter = read_filter.interpolate_filter()

        assert interp_filter.x.shape == (970,)
        assert interp_filter.y.shape == (970,)

        assert np.sum(interp_filter.x) == pytest.approx(1575.1079, rel=1e-8, abs=0.0)
        assert np.sum(interp_filter.y) == pytest.approx(514.135186, rel=1e-6, abs=0.0)

    def test_wavelength_range(self):
        read_filter = ReadFilter("MKO/NSFCam.H")
        min_wavel, max_wavel = read_filter.wavelength_range()

        assert min_wavel == pytest.approx(1.382, rel=1e-6, abs=0.0)
        assert max_wavel == pytest.approx(1.8656, rel=1e-8, abs=0.0)

    def test_mean_wavelength(self):
        read_filter = ReadFilter("MKO/NSFCam.H")
        mean_wavel = read_filter.mean_wavelength()

        assert mean_wavel == pytest.approx(1.6298258, rel=1e-6, abs=0.0)

    def test_filter_fwhm(self):
        read_filter = ReadFilter("MKO/NSFCam.H")
        filter_fwhm = read_filter.filter_fwhm()

        assert filter_fwhm == pytest.approx(0.2956945960962718, rel=self.limit, abs=0.0)



"""
Module for adding young, M- and L-type dwarf spectra from
`Allers & Liu (2013) <https://ui.adsabs.harvard.edu/abs/
2013ApJ...772...79A/abstract>`_ to the . These spectra are
also available in the `SpeX Prism Library Analysis Toolkit
<https://github.com/aburgasser/splat>`_.
"""

from pathlib import Path

import h5py
import numpy as np
import pandas as pd
import pooch

from astropy.io import fits
from astroquery.simbad import Simbad
from typeguard import typechecked

from species.util.data_util import extract_tarfile, remove_directory
from species.util.query_util import get_simbad


@typechecked
def add_allers2013(input_path: str, database: h5py._hl.files.File) -> None:
    """
    Function for adding the spectra of young, M- and L-type dwarfs
    from `Allers & Liu (2013) <https://ui.adsabs.harvard.edu/abs/
    2013ApJ...772...79A/abstract>`_  to the database.

    Parameters
    ----------
    input_path : str
        Path of the data folder.
    database : h5py._hl.files.File
        The HDF5 database.

    Returns
    -------
    NoneType
        None
    """

    Simbad.add_votable_fields("plx", "plx_error")

    url = "https://home.strw.leidenuniv.nl/~stolker/species/parallax.dat"
    input_file = "parallax.dat"
    data_file = Path(input_path) / input_file

    if not data_file.exists():
        print()

        pooch.retrieve(
            url=url,
            known_hash="e2fe0719a919dc98d24627a12f535862a107e473bc67f09298a40ad474cdd491",
            fname=input_file,
            path=input_path,
            progressbar=True,
        )

    parallax_data = pd.pandas.read_csv(
        data_file,
        usecols=[0, 1, 2],
        names=["object", "parallax", "parallax_error"],
        delimiter=",",
        dtype={"object": str, "parallax": float, "parallax_error": float},
    )

    print_text = "spectra of young M/L type objects from Allers & Liu 2013"

    url = "https://home.strw.leidenuniv.nl/~stolker/species/allers_liu_2013.tgz"
    input_file = "allers_liu_2013.tgz"
    data_file = Path(input_path) / input_file
    data_folder = Path(input_path) / "allers+2013/"

    if not data_file.exists():
        print()

        pooch.retrieve(
            url=url,
            known_hash=None,
            fname=input_file,
            path=input_path,
            progressbar=True,
        )

    if data_folder.exists():
        remove_directory(data_folder)

    print(f"\nUnpacking {print_text} (173 kB)...", end="", flush=True)
    extract_tarfile(str(data_file), str(data_folder))
    print(" [DONE]")

    data_file = Path(data_folder) / "sources.csv"

    sources = np.genfromtxt(
        data_file,
        delimiter=",",
        dtype=None,
        encoding="ASCII",
    )

    source_names = sources[:, 0]
    source_sptype = sources[:, 7]

    print_message = ""
    print()

    spec_files = sorted(data_folder.glob("*"))

    for file_item in spec_files:
        if file_item.suffix == ".fits":
            sp_data, header = fits.getdata(file_item, header=True)

        else:
            continue

        sp_data = np.transpose(sp_data)

        # (erg s-1 cm-2 A-1) -> (W m-2 um-1)
        sp_data[:, 1:] *= 10.0

        name = header["OBJECT"]

        if "RES" in header:
            spec_res = header["RES"]
        elif "RP" in header:
            spec_res = header["RP"]

        simbad_id = get_simbad(name)

        if simbad_id is not None:
            if not isinstance(simbad_id, str):
                simbad_id = simbad_id.decode("utf-8")

            par_select = parallax_data[parallax_data["object"] == simbad_id]

            if not par_select.empty:
                parallax = (
                    par_select["parallax"].values[0],
                    par_select["parallax_error"].values[0],
                )

            else:
                parallax = (np.nan, np.nan)

        else:
            parallax = (np.nan, np.nan)

        if np.isnan(parallax[0]) and simbad_id is not None:
            simbad_result = Simbad.query_object(simbad_id)

            if simbad_result is not None and len(simbad_result) > 0:
                if "PLX_VALUE" in simbad_result.columns:
                    if not simbad_result["PLX_VALUE"].mask[0]:
                        parallax = (
                            simbad_result["PLX_VALUE"].value[0],
                            simbad_result["PLX_ERROR"].value[0],
                        )

                else:
                    if not simbad_result["plx_value"].mask[0]:
                        parallax = (
                            simbad_result["plx_value"].value[0],
                            simbad_result["plx_err"].value[0],
                        )

        index = np.argwhere(source_names == name)

        if len(index) == 0:
            sptype = None
        else:
            sptype = source_sptype[index][0][0][:2]

        empty_message = len(print_message) * " "
        print(f"\r{empty_message}", end="")

        print_message = f"Adding spectra... {name}"
        print(f"\r{print_message}", end="")

        dset = database.create_dataset(f"spectra/allers+2013/{name}", data=sp_data)

        dset.attrs["name"] = str(name).encode()
        dset.attrs["sptype"] = str(sptype).encode()
        dset.attrs["simbad"] = str(simbad_id).encode()
        dset.attrs["parallax"] = float(parallax[0])  # (mas)
        dset.attrs["parallax_error"] = float(parallax[1])  # (mas)
        dset.attrs["spec_res"] = float(spec_res)

    empty_message = len(print_message) * " "
    print(f"\r{empty_message}", end="")

    print_message = "Adding spectra... [DONE]"
    print(f"\r{print_message}")



from species.core.species_init import SpeciesInit
from ._version import __version__

__author__ = "Tomas Stolker"
__license__ = "MIT"
__maintainer__ = "Tomas Stolker"
__email__ = "stolker@strw.leidenuniv.nl"
